//
// Copyright (c) Microsoft Corporation.  All rights reserved.
//
//
// Use of this sample source code is subject to the terms of the Microsoft
// license agreement under which you licensed this sample source code. If
// you did not accept the terms of the license agreement, you are not
// authorized to use this sample source code. For the terms of the license,
// please see the license agreement between you and Microsoft or, if applicable,
// see the LICENSE.RTF on your install media or the root of your tools installation.
// THE SAMPLE SOURCE CODE IS PROVIDED "AS IS", WITH NO WARRANTIES.
//
//------------------------------------------------------------------------------
//
//  File:  flash.c
//
//  Flash routines for the Intel Mainstone II bootloader.
//
#include <windows.h>
#include <mainstoneii.h>
#include <oal_memory.h>
#include <fmd.h>
#include "loader.h"

// Defines.
//
#define FLASH_DEVICE_ID_L18_MIN    0x880C
#define FLASH_DEVICE_ID_L18_MAX    0x8810
#define FLASH_DEVICE_ID_L30_MIN    0x8812
#define FLASH_DEVICE_ID_L30_MAX    0x8815
#define FLASH_DEVICE_ID_K3_MIN     0x8801
#define FLASH_DEVICE_ID_K3_MAX     0x8803
#define FLASH_DEVICE_ID_K18_MIN    0x8805
#define FLASH_DEVICE_ID_K18_MAX    0x8807
#define FLASH_DEVICE_ID_J3_MIN     0x0016
#define FLASH_DEVICE_ID_J3_MAX     0x0018
#define FLASH_DEVICE_ID_P30_MIN    0x881a
#define FLASH_DEVICE_ID_P30_MAX    0x891c

// Type definitions.
//
#define FIOW	UINT16		// flash io width

typedef enum _FLASH_BLOCK_ARRAY_TYPE_
{
    CONSTANT_SIZE,
    BOTTOM_PARAMETER,
    TOP_PARAMETER
} FLASH_BLOCK_ARRAY_TYPE, *PFLASH_BLOC_ARRAY_TYPE;

typedef enum _FLASH_DEVICE_TYPE_
{
    NONE,
    L18, 
    L30,
    K3,
    K18,
    J3,
	P30
} FLASH_DEVICE_TYPE, *PFLASH_DEVICE_TYPE;

// Local variables.
//
static FLASH_BLOCK_ARRAY_TYPE g_BlockArrayType  = CONSTANT_SIZE;
static FLASH_DEVICE_TYPE      g_FlashDeviceType = NONE;

//#ifdef MCPII_MODULE
//static UINT32 FLASH_SIZE    = 0x04000000;       // 64MB.
//static UINT32 ERASE_BLOCKS  = 256;              // Total number of blocks (hiding parameter blocks).
//#else
static UINT32 FLASH_SIZE    = 0x02000000;       // 32MB.
static UINT32 ERASE_BLOCKS  = 256;              // Total number of blocks (hiding parameter blocks).
//#endif
static UINT32 L3_BLOCK_SIZE = 0x10000;          // 64KB per block on L3 (blocks 0 - 3).

// Global variables.
//
extern IMAGE_TYPE g_ImageType;

extern UINT16 g_UniqueDeviceID[4];


//------------------------------------------------------------------------------
//
//  Function:  OEMIsFlashAddr
//
// Checks the virtual address provided and determines if it's a flash address or not.
//
BOOL OEMIsFlashAddr(DWORD dwAddr)
{
    DWORD dwPhysAddr = (DWORD) OALVAtoPA((void*) dwAddr);
    
    if ((dwPhysAddr >= MAINSTONEII_BASE_PA_BOOT_FLASH) && (dwPhysAddr < (MAINSTONEII_BASE_PA_BOOT_FLASH + MAINSTONEII_SIZE_BOOT_FLASH )))
    {
        return(TRUE);
    }
    
    return(FALSE);
}


//------------------------------------------------------------------------------
//
//  Function:  OEMMapMemAddr
//
// Maps a flash address to a RAM address (used when temporarily caching a
// flash-destined download image to RAM).
//
LPBYTE OEMMapMemAddr (DWORD dwImageStart, DWORD dwAddr)
{

    if (OEMIsFlashAddr(dwAddr))
    {
        // The image being downloaded is a flash image - temporarily
        // cache the image in RAM until it's completely downloaded.
        //
        dwAddr -= dwImageStart;
        dwAddr += 0x80100000;
    }

    return((LPBYTE) (dwAddr));
}


//------------------------------------------------------------------------------
//
//  Function:  OEMStartEraseFlash
//
// Begins the flash erase procedure.
//
BOOL OEMStartEraseFlash (DWORD dwStartAddr, DWORD dwLength)
{
    return(TRUE);
}


//------------------------------------------------------------------------------
//
//  Function:  main
//
// Continues the flash erase procedure (erases are incrementally done during
// download to speed up the overall flash operation from the user's perspective).
//
void OEMContinueEraseFlash(void)
{
}


//------------------------------------------------------------------------------
//
//  Function:  OEMFinishEraseFlash
//
// Finished the flash erase procedure.
//
BOOL OEMFinishEraseFlash(void)
{
    return(TRUE);
}


BOOL OEMWriteFlash(DWORD dwImageStart, DWORD dwImageLength)
{
    BOOL  retVal = FALSE;
    UINT8 *pDataBuffer = (UINT8 *) OEMMapMemAddr(dwImageStart, dwImageStart);

    // FlashWrite uses virtual uncached addresses
    // dwImageStart could be either virtual uncached (eboot.bin) or virtual cached (nk.bin)
    dwImageStart = (DWORD) OALCAtoUA(dwImageStart);

    retVal = FlashWrite(dwImageStart, pDataBuffer, dwImageLength);

    // If we just wrote the bootloader to flash, halt and require the system to be rebooted.
    if (g_ImageType == IMAGE_TYPE_BOOTLOADER)
    {
        KITLOutputDebugString("INFO: bootloader image stored successfully.\r\n");
        KITLOutputDebugString("Reboot the system manually.  ");
        SpinForever();
    }

    return(retVal);

}


// START: these routines will be replaced by the FMD when it's done.
static FLASH_DEVICE_TYPE DetectFlashDevice(UINT32 FlashBaseAddress)
{
    UINT32 nDeviceID = 0;
    volatile FIOW *pFlash = (volatile FIOW *)(FlashBaseAddress);

    // Get the flash part ID.
    //
    *pFlash   = (FIOW) 0x00900090;
    nDeviceID = (*(pFlash + 1) & 0xFFFF);

    // Get the flash part Unique ID and store for later use
    g_UniqueDeviceID[0] = *(pFlash + 0x81);
    g_UniqueDeviceID[1] = *(pFlash + 0x82);
    g_UniqueDeviceID[2] = *(pFlash + 0x83);
    g_UniqueDeviceID[3] = *(pFlash + 0x84);

    // Put the flash part in read array mode.
    //
    *pFlash   = (FIOW) 0x00FF00FF;

    if ((nDeviceID >= FLASH_DEVICE_ID_L18_MIN) && (nDeviceID <= FLASH_DEVICE_ID_L18_MAX))
    {
        //KITLOutputDebugString("INFO: DetectFlashDevice: flash type is L18.\r\n");
        return(L18);
    }
    if ((nDeviceID >= FLASH_DEVICE_ID_L30_MIN) && (nDeviceID <= FLASH_DEVICE_ID_L30_MAX))
    {
        //KITLOutputDebugString("INFO: DetectFlashDevice: flash type is L30.\r\n");
        return(L30);
    }
    if ((nDeviceID >= FLASH_DEVICE_ID_K3_MIN) && (nDeviceID <= FLASH_DEVICE_ID_K3_MAX))
    {
        //KITLOutputDebugString("INFO: DetectFlashDevice: flash type is K3.\r\n");
        return(K3);
    }
    if ((nDeviceID >= FLASH_DEVICE_ID_K18_MIN) && (nDeviceID <= FLASH_DEVICE_ID_K18_MAX))
    {
        //KITLOutputDebugString("INFO: DetectFlashDevice: flash type is K18.\r\n");
        return(K18);
    }
    if ((nDeviceID >= FLASH_DEVICE_ID_J3_MIN) && (nDeviceID <= FLASH_DEVICE_ID_J3_MAX))
    {
        //KITLOutputDebugString("INFO: DetectFlashDevice: flash type is J3.\r\n");
        return(J3);
    }
    if ((nDeviceID >= FLASH_DEVICE_ID_P30_MIN) && (nDeviceID <= FLASH_DEVICE_ID_P30_MAX))
    {
//        KITLOutputDebugString("INFO: DetectFlashDevice: flash type is P30.\r\n");
        return(P30);
    }

    KITLOutputDebugString("INFO: DetectFlashDevice: flash type is undetermined.\r\n");
    return(NONE);
}


static BOOL FlashErase(UINT32 FlashStart, UINT32 FlashLength)
{
    UINT32 j;
    UINT32 num_blocks;
    UINT32 num_blocks_to_erase;
    UINT32 num_l3_blocks_to_erase = 0;
    UINT32 num_l3_blocks_erased;
    UINT32 flash_start_address = (UINT32) OALPAtoVA(MAINSTONEII_BASE_PA_BOOT_FLASH, FALSE);
    volatile FIOW *pFlash;
    FIOW status;
    UINT32 BLOCK_SIZE = (FLASH_SIZE / ERASE_BLOCKS);

    // Make sure we know what kind of flash part we have.
    //
    if (g_FlashDeviceType == NONE)
    {
        KITLOutputDebugString("ERROR: FlashErase: unrecognized flash part.\r\n");
        return(FALSE);
    }

    FlashStart = (UINT32) OALCAtoUA(FlashStart);
    // Determine the number of blocks to erase.
    num_blocks = (FlashLength / (BLOCK_SIZE));
    if (FlashLength % (BLOCK_SIZE))
    {
        num_blocks++;
    }

    if (num_blocks < 1)
    {
        num_blocks = 1;
        KITLOutputDebugString("WARNING: FlashErase: calculation error.  Erase blocks = %d\n\r", num_blocks);
    }

    else if (num_blocks > ERASE_BLOCKS)
    {
        num_blocks = ERASE_BLOCKS;
        KITLOutputDebugString("WARNING: FlashErase: calculation error.  Erase blocks = %d\n\r", num_blocks);
    }

    // If flash, need to change block count based on if we're in the first 4 blocks
    // as the first 4 blocks are 16KWords (vs. 64KWords/block)
    if ((g_FlashDeviceType == L18) || (g_FlashDeviceType == L30))
    {
        // Is start address within the first 4 blocks.  As there are 2 devices in parallel,
        // are we within the first 128KWords (2 x 64KWords)?  If so, we need to account for
        // increased number of blocks - which is the first 256K of addy range
        if ((FlashStart >= flash_start_address) && (FlashStart < (flash_start_address + 0x40000)))
        {
            // If we're within the first 128K words (256K), we need to treat the block as a special case
            num_l3_blocks_to_erase = 4;
            num_blocks--; // Decrease the num_blocks count, as it assumes the first block is 64K

            // Now add the # of L3 blocks to the # blocks calc.'ed at the beginning to get total
            // # of blocks to erase on a given L3 device INCLUDING the first 4 16kWord blocks
            num_blocks = num_blocks + num_l3_blocks_to_erase;
        }
    }

    pFlash = (volatile FIOW *)(FlashStart);

    // Issue the clear lock bits and confirm command.
    if (g_FlashDeviceType == J3)
    {
        // For J3 FLASH, unlock all blocks at once with one command
        num_blocks_to_erase = 1;
    }
    else if ((g_FlashDeviceType == K3) || (g_FlashDeviceType == K18)||
		(g_FlashDeviceType == L18) || (g_FlashDeviceType == L30) || (g_FlashDeviceType == P30))
    {
        // For K3/K18 FLASH, unlock individual blocks one at a time
        num_blocks_to_erase = num_blocks;
    }

    // For L3, need to set for use within the FOR loop
    num_l3_blocks_erased = num_l3_blocks_to_erase;

    for (j = 0; j < num_blocks_to_erase; j++)
    {
        *pFlash = (FIOW) 0x00600060;
        *pFlash = (FIOW) 0x00d000d0;
        while ((*pFlash & 0x00800080) != (FIOW) 0x00800080);

        status = *pFlash;
        if ((status & 0x00000008) == 0x00000008)
		{
            KITLOutputDebugString("ERROR: FlashErase: voltage range error ... lower flash.\r\n");
		}
        if ((status & 0x00080000) == 0x00080000)
		{
            KITLOutputDebugString("ERROR: FlashErase: voltage range error ... upper flash.\r\n");
		}
        if ((status & 0x00000030) == 0x00000030)
		{
            KITLOutputDebugString("ERROR: FlashErase: command sequence error ... lower flash.\r\n");
		}
        if ((status & 0x00300000) == 0x00300000)
		{
            KITLOutputDebugString("ERROR: FlashErase: command sequence error ... upper flash.\r\n");
		}
        if ((status & 0x00000020) == 0x00000020)
		{
            KITLOutputDebugString("ERROR: FlashErase: clear lock bits error ... lower flash.\r\n");
		}
        if ((status & 0x00200000) == 0x00200000)
		{
            KITLOutputDebugString("ERROR: FlashErase: clear lock bits error ... upper flash.\r\n");
		}

        if (status != (FIOW) 0x00800080)
        {
            KITLOutputDebugString("ERROR: FlashErase: status register returned 0x%X\r\n", status);
            KITLOutputDebugString("ERROR: FlashErase: unrecoverable failure encountered while erasing flash.  System halted!\r\n");
            return(FALSE);
        }
        if ( ((g_FlashDeviceType == L18) || (g_FlashDeviceType == L30)) && (num_l3_blocks_erased > 0) )
        {
            // We need to take care of the first 16K blocks if there are any
            pFlash += (L3_BLOCK_SIZE / 4);
            num_l3_blocks_erased--;
        }
        else
        {
            pFlash += (BLOCK_SIZE / sizeof(FIOW));
        }
    }
    pFlash = (volatile FIOW *)(FlashStart);

    // Clear the status register
    *pFlash = (FIOW) 0x00500050;

    if ( ((g_FlashDeviceType == L18) || (g_FlashDeviceType == L30)) && (num_l3_blocks_to_erase > 0) )
    {
        KITLOutputDebugString("INFO: FlashErase: erasing flash %X to %X.  Please wait... \r\n", FlashStart, FlashStart + (((num_blocks-num_l3_blocks_to_erase)*256*1024) + (num_l3_blocks_to_erase*64*1024)) - 1);
    }
    else
    {
        KITLOutputDebugString("INFO: FlashErase: erasing flash %X to %X.  Please wait... \r\n", FlashStart, FlashStart + (num_blocks*256*1024) - 1);
    }

    // For L3, need to set for use within the FOR loop
    num_l3_blocks_erased = num_l3_blocks_to_erase;

    // Erase each block.
    for (j = 0; j < num_blocks; j++)
    {
        // Issue erase and confirm command
        *pFlash = (FIOW) 0x00200020;
        *pFlash = (FIOW) 0x00d000d0;
        while ((*pFlash & 0x00800080) != (FIOW) 0x00800080);

        // Check if the flash block erased successfully.
        // Was either block locked?
        status = *pFlash;
        if ((status & 0x00200000) || (status & 0x00000020))
        {
            if (status & 0x00200000)
            {
                KITLOutputDebugString("ERROR: FlashErase: block erase failure.  Lock bit upper flash set!\r\n");
            }

            if (status & 0x00000020)
            {
                KITLOutputDebugString("ERROR: FlashErase: block erase failure.  Lock bit lower flash set!\r\n");
            }

            KITLOutputDebugString("ERROR: FlashErase: unrecoverable failure encountered while erasing flash.\r\n");
            return(FALSE);
        }
        KITLOutputDebugString(".");

        if ( ((g_FlashDeviceType == L18) || (g_FlashDeviceType == L30)) && (num_l3_blocks_erased > 0) )
        {
            pFlash += (L3_BLOCK_SIZE / sizeof(FIOW));
            num_l3_blocks_erased--;
        }
        else
        {
            pFlash += (BLOCK_SIZE / sizeof(FIOW));
        }
    }

    KITLOutputDebugString("\r\n");
    pFlash = (volatile FIOW *)(FlashStart);

    // Put the flash back into read mode
    if ((g_FlashDeviceType == L18) || (g_FlashDeviceType == L30))
    {
        // Setting the number of 16K blocks that we have to deal with
        num_l3_blocks_erased = num_l3_blocks_to_erase;

        // Put each block back into read mode
        for (j = 0; j < num_blocks; j++)
        {
            *pFlash = (FIOW) 0x00FF00FF;

            if ( num_l3_blocks_erased )
            {
                pFlash += (L3_BLOCK_SIZE / sizeof(FIOW));
                num_l3_blocks_erased--;
            }
            else
            {
                pFlash += (BLOCK_SIZE / sizeof(FIOW));
            }
        }
        // Now reset the flash pointer
        pFlash = (volatile FIOW *)(FlashStart);
    }
    else
    {
        *pFlash = (FIOW) 0x00FF00FF;
    }

    // Flash erase verification.
    for (j = 0; j < (FlashLength / sizeof(FIOW)); j++)
    {
        if (*pFlash != (FIOW) 0xFFFFFFFF)
        {
            KITLOutputDebugString("ERROR: FlashErase: erase verification failure at address 0x%X Data 0x%X.\r\n", pFlash, *pFlash);
            KITLOutputDebugString("ERROR: FlashErase: unrecoverable failure encountered while erasing flash.\r\n");
            return(FALSE);
        }
		pFlash++;
    }

    return(TRUE);
}


BOOL FlashInit(UINT32 FlashStart, UINT32 FlashLength)
{
    FlashStart = (UINT32) OALCAtoUA(FlashStart);

    // Read the device ID and determine the flash part type.
    //
    g_FlashDeviceType = DetectFlashDevice(FlashStart);

    return((g_FlashDeviceType != NONE) ? TRUE : FALSE);
}


BOOL FlashRead(UINT32 FlashStart, UINT8 *pDataBuffer, UINT32 DataLength)
{
    if (!pDataBuffer) return(FALSE);

    FlashStart = (UINT32) OALCAtoUA(FlashStart);

    // TODO - check flash address...
    memcpy((void *)pDataBuffer, (const void *)FlashStart, DataLength);

    return(TRUE);
}


// TODO - block-aligned address.
BOOL FlashWrite(UINT32 FlashStart, UINT8 *pDataBuffer, UINT32 DataLength)
{
    UINT32 j,b;
    UINT32 sizeFlashCache; // size in bytes of the flash cache
    UINT32 chunksPerBlock;
    UINT32 count;
    UINT32 val = 0;
    UINT32 dwLength = DataLength;

    volatile FIOW *pFlash         = (volatile FIOW *) FlashStart;
    volatile FIOW *pBlockAddress  = (volatile FIOW *) FlashStart;
    volatile FIOW *pDeviceAddress = (volatile FIOW *) FlashStart;
    volatile FIOW *pFlashCache    = (FIOW *) pDataBuffer;

    volatile FIOW *pdwFlashStart;
    volatile FIOW *pdwRamStart;


    if (g_FlashDeviceType == NONE)
    {
        KITLOutputDebugString("ERROR: FlashWrite: unrecognized flash part.\r\n");
        return(FALSE);
    }

    if (!FlashErase(FlashStart, DataLength))
    {
        return(FALSE);
    }
        
    b = 0;

    if (g_FlashDeviceType == J3)
    {
        sizeFlashCache = 64;   // 64 bytes total, 32 bytes per device
        chunksPerBlock = 4096; // 256K / 64bytes = 4096
    }
    else if ((g_FlashDeviceType == K3) || (g_FlashDeviceType == K18) ||
		(g_FlashDeviceType == L18) || (g_FlashDeviceType == L30))
    {
        sizeFlashCache = 128;  // 128 bytes total, 64 bytes per device
        chunksPerBlock = 2048; // 256K / 128bytes = 2048
    }
	else
	{
        sizeFlashCache = 64;  // 64 bytes total, 64 bytes per device
        chunksPerBlock = 2048; // 128K / 64bytes = 2048
	}

    KITLOutputDebugString("INFO: FlashWrite: writing to flash...\r\n");

    while (DataLength > 0)
    {
        // Calculate the number of DWORDS to program in each iteration based on the size in bytes
        // of the flash cache.
        if (DataLength > sizeFlashCache)
            count = sizeFlashCache / sizeof(FIOW);
        else
            count = DataLength / sizeof(FIOW);

        // Issue Write to Buffer Command
        *pBlockAddress = (FIOW) 0x00E800E8;

        // Check that the write buffer is available.
        // Read the Extended Status Register and
        // wait for the assertion of XSR[7] before continuing.
        while ( (*pBlockAddress & 0x00800080) != (FIOW) 0x00800080)
        {
            *pBlockAddress = (FIOW) 0x00E800E8;
        }

        // Configure the size of the buffer that we will write.
        // Write the word count (device expects a 0 based number)
        *pBlockAddress = (FIOW) ((count - 1) << 16 | (count - 1));

        pdwFlashStart = pDeviceAddress;
        pdwRamStart   = pFlashCache;

        // Write up to "count" DWORDS into the Flash memory staging area
        for (j = 0; j < count; j++)
        {
            *pDeviceAddress++ = *pFlashCache++;
        }

        // increment the number of 64 or 128 byte segments written.
        b++;
        // Now program the buffer into Flash
        *pBlockAddress = (FIOW) 0x00D000D0;
        while ((*pBlockAddress & 0x00800080) != (FIOW) 0x00800080);

        DataLength -= (count * sizeof(FIOW));

        // Read back the segment just written
        *pFlash = (FIOW) 0x00FF00FF;
        for (j = 0; j < count; j++)
        {
            if ((g_FlashDeviceType == L18) || (g_FlashDeviceType == L30))
            {
                *pdwFlashStart = (FIOW) 0x00FF00FF;
            }

            if (*pdwFlashStart != *pdwRamStart)
            {
                KITLOutputDebugString("ERROR: FlashWrite: flash programming failure at address 0x%x (expected 0x%x actual 0x%x).\r\n", pdwFlashStart, *pdwRamStart, *pdwFlashStart);
                return(FALSE);
            }
			pdwFlashStart++;
			pdwRamStart++;
        }

        if ((b % chunksPerBlock) == 0)
        {
            KITLOutputDebugString(".");
        }
    }

    // Verify the data written to flash...
    //
    pFlash      = (volatile FIOW *) FlashStart;
    pFlashCache = (FIOW *) pDataBuffer;

    // Put Flash into read mode.
    *pFlash = (FIOW) 0x00FF00FF;

    KITLOutputDebugString("\r\nINFO: FlashWrite: verifying the data written to flash...\r\n");

    for ( j = 0; j < (dwLength/sizeof(FIOW)); j++ )
    {
        if (*pFlash != *pFlashCache)
        {
            KITLOutputDebugString( "ERROR: FlashWrite: verification failure at address 0x%x (expected 0x%x actual 0x%x))\r\n", pFlash, *pFlashCache, *pFlash );
            return(FALSE);
        }
		pFlash++;
		pFlashCache++;
    }

    KITLOutputDebugString("INFO: FlashWrite: flash programmed successfully!\r\n");

    return(TRUE);
}

// END: these routines will be replaced by the FMD when it's done.
