//
// Copyright (c) Microsoft Corporation.  All rights reserved.
//
//
// Use of this sample source code is subject to the terms of the Microsoft
// license agreement under which you licensed this sample source code. If
// you did not accept the terms of the license agreement, you are not
// authorized to use this sample source code. For the terms of the license,
// please see the license agreement between you and Microsoft or, if applicable,
// see the LICENSE.RTF on your install media or the root of your tools installation.
// THE SAMPLE SOURCE CODE IS PROVIDED "AS IS", WITH NO WARRANTIES.
//
//------------------------------------------------------------------------------
//
//  File:  main.c
//
//  Core routines for the Intel Mainstone bootloader.
//
#include <windows.h>
#include <nkintr.h>
#include <bulverde.h>
#include <mainstoneii.h>
#include <oal_memory.h>
#include <pcireg.h>
#include <fmd.h>
#include <xllp_pccardsocket.h>
#include <bsp.h>
#include <ethdrv.h>
#include "loader.h"

#define NUM_BOOT_DEVICES  5
#define NUM_BOOT_ORDERS   6
#define SDBG_USB_SERIAL   101

//------------------------------------------------------------------------------
// Local variables.
//
static PCI_REG_INFO g_FlashAddress;
EBOOT_CFG           g_EbootCFG;
static BOOLEAN      g_DownloadImage = TRUE;
BOOLEAN             g_SerialUSBDownload = FALSE;

// This table describes the boot order for a given configuration.
//
static BOOT_DEVICE_TYPE BootOrder[NUM_BOOT_ORDERS][NUM_BOOT_DEVICES] = {
    { ETH_DEVICE_DM9000,    ETH_DEVICE_USB,     ETH_DEVICE_PCMCIA0, ETH_DEVICE_PCMCIA1, SER_DEVICE_USB },  // pEbootCFG->bootDeviceOrder = 0.
    { ETH_DEVICE_PCMCIA0, ETH_DEVICE_PCMCIA1, ETH_DEVICE_DM9000,    ETH_DEVICE_USB, SER_DEVICE_USB  },       // pEbootCFG->bootDeviceOrder = 1.
    { ETH_DEVICE_PCMCIA1, ETH_DEVICE_PCMCIA0, ETH_DEVICE_DM9000,    ETH_DEVICE_USB, SER_DEVICE_USB },       // pEbootCFG->bootDeviceOrder = 2.
    { ETH_DEVICE_USB,     ETH_DEVICE_DM9000,    ETH_DEVICE_PCMCIA0, ETH_DEVICE_PCMCIA1, SER_DEVICE_USB },  // pEbootCFG->bootDeviceOrder = 3.
    { SER_DEVICE_USB,     ETH_DEVICE_DM9000,    ETH_DEVICE_USB,     ETH_DEVICE_PCMCIA0, ETH_DEVICE_PCMCIA1 },  // pEbootCFG->bootDeviceOrder = 4.
    { SER_DEVICE_USB,     ETH_DEVICE_USB,     ETH_DEVICE_DM9000,    ETH_DEVICE_PCMCIA0, ETH_DEVICE_PCMCIA1 },  // pEbootCFG->bootDeviceOrder = 5.
};

static char* BootOrderString[NUM_BOOT_ORDERS] =
{
    "DM9000, USB RNDIS, PCMCIA0, PCMCIA1, USB SERIAL (DL Only)",
    "PCMCIA0, PCMCIA1, DM9000, USB RNDIS, USB SERIAL (DL Only)",
    "PCMCIA1, PCMCIA0, DM9000, USB RNDIS, USB SERIAL (DL Only)",
    "USB RNDIS, DM9000, PCMCIA0, PCMCIA1, USB SERIAL (DL Only)",
    "USB SERIAL (DL Only), DM9000, USB RNDIS, PCMCIA0, PCMCIA1",
    "USB SERIAL (DL Only), USB RNDIS, DM9000, PCMCIA0, PCMCIA1",
};

//------------------------------------------------------------------------------
// Global variables.
//
FlashInfo g_FlashInfo;
EDBG_ADDR g_DeviceAddr; // NOTE: global used so it remains in scope throughout download process
                        // since eboot library code keeps a global pointer to the variable provided.
XLLP_PCCARDSOCKET_T strEbtPCCardSocketHandle;
IMAGE_TYPE g_ImageType;

extern BSP_ARGS *g_pBSPArgs;

//------------------------------------------------------------------------------
// Local function prototypes.
//
static BOOL LoadEBootCFG(EBOOT_CFG *EBootCFG);
static BOOL StoreEBootCFG(EBOOT_CFG *EBootCFG);
static void ResetDefaultEBootCFG(EBOOT_CFG *pEbootCFG);

static void SetMAC(unsigned char *mac);
static void SetIP(EBOOT_CFG *pEbootCFG);
static void SetMask(EBOOT_CFG *pEbootCFG);
static void SetBootMe(EBOOT_CFG *pEbootCFG);
static void SetDelay(EBOOT_CFG *pEbootCFG);

BOOL OEMVerifyMemory(DWORD dwStartAddr, DWORD dwLength);

//------------------------------------------------------------------------------
// External function prorotypes.
//
extern void  Launch(unsigned int uAddr);
extern BOOL  SerialSendBlockAck(DWORD uBlockNumber);
extern BOOL  SerialSendBootRequest(const char * platformString);
extern BOOL  SerialWaitForBootAck(BOOL *pfJump);
extern DWORD SerialWaitForJump(VOID);

extern BOOL USBSerialInit();


//------------------------------------------------------------------------------
//
//  Function:  main
//
//  Bootloader main routine.
//
void main(void)
{

    // Common boot loader (blcommon) main routine.
    //
    BootloaderMain();

    // Should never get here.
    //
    SpinForever();

}


//------------------------------------------------------------------------------
//
//  Function:  OEMDownloadFileNotify
//
//  This routine receives a download manifest from blcommon describing the .bin
//  and/or .nb0 file(s) being downloaded from the host.  It's purpose is to
//  assign an address (typically flash address) where the .nb0 file(s) should be
//  placed (since .nb0 files aren't self-describing like .bin files are) and to
//  display the manifest information to the user (over serial).
//
void OEMDownloadFileNotify(PDownloadManifest pInfo)
{
    DWORD dwCount;
    DWORD dwNumNB0Files = 0;

    if (!pInfo) return;

    KITLOutputDebugString("\r\nDownload file information:\r\n");
    KITLOutputDebugString("-------------------------------------------------------------------------------\r\n");

    for (dwCount = 0 ; dwCount < pInfo->dwNumRegions ; dwCount++)
    {

        KITLOutputDebugString("[%d]: Address=0x%x  Length=0x%x  Name=%s\r\n", dwCount, 
                              pInfo->Region[dwCount].dwRegionStart, 
                              pInfo->Region[dwCount].dwRegionLength, 
                              pInfo->Region[dwCount].szFileName);

        // .nb0 files will have a start address of 0 because Platform Builder
        // won't know where to place them.  We'll support only one .nb0 file
        // download (this is an Image Update disk image).  If we need to
        // support more than one .nb0 file download in the future, we'll need
        // to differentiate them by filename.
        //
        if (pInfo->Region[dwCount].dwRegionStart == 0)
        {
            // We only support one raw binary file (disk image).
            if (dwNumNB0Files++)
            {
                KITLOutputDebugString("ERROR: This bootloader doesn't support downloading a second .nb0 binary image.\r\n");
                SpinForever();
            }

            // The disk image .nb0 file should be placed immediately after the
            // bootloader image in flash (we'll assume that the disk image has the IPL prepended to it).
            pInfo->Region[dwCount].dwRegionStart = (DWORD)OALPAtoVA(IMAGE_BOOT_IPL_FLASH_PA_START, FALSE);

            KITLOutputDebugString("INFO: This bootloader assumes any .nb0 file is an Image Update disk image.\r\n");
            KITLOutputDebugString("INFO: Storing file \"%s\" in flash after boot loader.\r\n", pInfo->Region[dwCount].szFileName);
            KITLOutputDebugString("INFO: Changed start address for %s to 0x%x.\r\n", pInfo->Region[dwCount].szFileName, 
                                                                                     pInfo->Region[dwCount].dwRegionStart);

        }
    }

    KITLOutputDebugString("\r\n");

    return;
}


//------------------------------------------------------------------------------
//
//  Function:  OEMDebugInit
//
//  Initialize debug serial UART.
//
BOOL OEMDebugInit(void)
{
    // Initialize the flash interface (needed so we can determine which serial
    // port to use for bootloader and OS debug message output).
    //
    if (!FlashInit((UINT32) OALPAtoVA(MAINSTONEII_BASE_PA_BOOT_FLASH, FALSE), MAINSTONEII_SIZE_BOOT_FLASH))
    {
        // Load default bootloader configuration settings.
        //
        KITLOutputDebugString("ERROR: flash initialization failed - loading bootloader defaults...\r\n");
        ResetDefaultEBootCFG(&g_EbootCFG);
    }
    else
    {
        // Load the bootloader configuration from flash (menu settings).
        //
        LoadEBootCFG(&g_EbootCFG);
    }

    // Check and initialize the BSP Args area
    // Note: OALArgsInit should be called AFTER FlashInit so that 
    // the g_UniqueDeviceID structure can be populated. This is to ensure
    // we have unique device data for UUID's, etc.
    OALArgsInit((BSP_ARGS *) IMAGE_SHARE_ARGS_UA_START);

    // Initialize the debug UART.
    //
    InitDebugSerial(g_EbootCFG.dwDbgSerPhysAddr, FALSE);

    // Clear the hex LEDs.
    //
    OEMWriteDebugLED(0, 0);

    // Set up optional bootloader function pointers.
    //
    g_pOEMVerifyMemory   = OEMVerifyMemory;
    g_pOEMMultiBINNotify = OEMDownloadFileNotify;

    return(TRUE);
}


//------------------------------------------------------------------------------
//
//  Function:  OEMPlatformInit
//
//  Initialize the Mainstone platform.
//
BOOL OEMPlatformInit(void)
{
    UINT32 AutoBootDelay = 0;
    BOOLEAN bXIPMode     = TRUE;
    BOOLEAN bCFGChanged  = FALSE;
    SYSTEMTIME SysTime;
    UINT32 StartTime, CurrTime, PrevTime;
    UINT32 Selection;
    UINT32 BootDevice;
    int i;
    UCHAR  rndis_mac[6] = {0,0,0,0,0,0};

    KITLOutputDebugString("Microsoft Windows CE Ethernet Bootloader %d.%d for the Voipac PXA270 Development Platform Built %s\r\n", \
                          EBOOT_VERSION_MAJOR, EBOOT_VERSION_MINOR, __DATE__);

    // Initialize the global XLLP device handle for the PC Card interface.
    //
    memset((void *) &strEbtPCCardSocketHandle, 0, sizeof(XLLP_PCCARDSOCKET_T));
    strEbtPCCardSocketHandle.pstrGpioRegsHandle = (XLLP_GPIO_T *) OALPAtoVA(BULVERDE_BASE_REG_PA_GPIO, FALSE);
//    XllpPCCardConfigureGPIOs(&strEbtPCCardSocketHandle);

    // Get the Ehternet MAC address to use as RNDIS MAC address
    // Note: This only applies to Mainstone platform because it has Ehternet controller on it.
    //
    if (!g_EbootCFG.RNDISMac[0] && !g_EbootCFG.RNDISMac[1] && !g_EbootCFG.RNDISMac[2])
    {
        if (DM9000Init((UINT8 *) (((UINT32) OALPAtoVA(MAINSTONEII_BASE_REG_PA_DM9000_ETHERNET, FALSE)) + 0x300), 0, g_EbootCFG.RNDISMac))
        {
            bCFGChanged = TRUE;
        }
    }

    // Read the current wall-clock time.
    // NOTE: At reset, the RTC value is set to 1/1/1980 at 0:0:0.
    //
    OEMGetRealTime(&SysTime);

    // User menu code...
    //
    AutoBootDelay = g_EbootCFG.delay;
    if (g_EbootCFG.autoDownloadImage)
    {
        g_DownloadImage = TRUE;
        KITLOutputDebugString ( "\r\nPress [ENTER] to download now or [SPACE] to cancel.\r\n");
        KITLOutputDebugString ( "\r\nInitiating image download in %d seconds. ", AutoBootDelay--);
    }
    else
    {
        g_DownloadImage = FALSE;
        KITLOutputDebugString ( "\r\nPress [ENTER] to launch image stored in flash or [SPACE] to cancel.\r\n");
        KITLOutputDebugString ( "\r\nInitiating image launch in %d seconds. ", AutoBootDelay--);
    }

    // Get a snapshot of the RTC seconds count.
    //
    StartTime     = OEMKitlGetSecs();
    PrevTime      = StartTime;
    CurrTime      = StartTime;
    Selection     = 0;

    // Allow the user an amount of time to halt the auto boot/download process.
    // Count down to 0 before proceeding with default operation.
    //
    while ((CurrTime - StartTime) < g_EbootCFG.delay)
    {
        UINT8 k=0;
        UINT8 j;
        UINT8 x,y,z;

        Selection = OEMReadDebugByte(); 
        if ((Selection == 0x20) || (Selection == 0x0d))
        {
            break;
        }
        CurrTime = OEMKitlGetSecs();   
        if (CurrTime > PrevTime)
        {
            PrevTime = CurrTime;
            if (AutoBootDelay < 9)
                k = 11;
            else if (AutoBootDelay < 99)
                k = 12;
            else if (AutoBootDelay < 999)
                k = 13;

            for (j = 0; j < k; j++)
            {
                OEMWriteDebugByte((BYTE)0x08); // print back space
            }
            x = AutoBootDelay / 100;
            y = (AutoBootDelay % 100) / 10;
            z = ((AutoBootDelay % 100) % 10);
            OEMWriteDebugLED(0, ((x << 8) | (y << 4) | (z)));
            KITLOutputDebugString ( "%d seconds. ", AutoBootDelay--);
        }

    }

    switch (Selection)
    {
    case 0x00: // fall through if nothing typed
    case 0x0d: // user canceled wait
        {
            if (g_EbootCFG.autoDownloadImage)
            {
                KITLOutputDebugString ( "\r\nStarting auto download ... \r\n");
            }
            else
            {
                KITLOutputDebugString ( "\r\nLaunching flash image  ... \r\n");
            }
            break;
        }
    case 0x20:
        {
            Selection = 0;
            while (1)
            {
                // Get the Ehternet MAC address to use as RNDIS MAC address
                // Note: This only applies to Mainstone platform because it has Ehternet controller on it.
                //
                if (!g_EbootCFG.RNDISMac[0] && !g_EbootCFG.RNDISMac[1] && !g_EbootCFG.RNDISMac[2])
                {
                    if (DM9000Init((UINT8 *) (((UINT32) OALPAtoVA(MAINSTONEII_BASE_REG_PA_DM9000_ETHERNET, FALSE)) + 0x300), 0, g_EbootCFG.RNDISMac))
                    {
                        bCFGChanged = TRUE;
                    }
                }

                // Show menu
                KITLOutputDebugString ( "\r\n\r\nEBoot Loader Configuration:\r\n\r\n");

                KITLOutputDebugString ( "0) IP address: %s\r\n",inet_ntoa(g_EbootCFG.IP));
                KITLOutputDebugString ( "1) Subnet mask: %s\r\n", inet_ntoa(g_EbootCFG.subnetMask));
                KITLOutputDebugString ( "2) Boot delay: %d seconds\r\n", g_EbootCFG.delay);
                KITLOutputDebugString ( "3) DHCP: %s\r\n", (g_EbootCFG.DHCPEnable == TRUE?"(Enabled)":"(Disabled)"));
                KITLOutputDebugString ( "4) Reset to factory default configuration\r\n");
                KITLOutputDebugString ( "5) RNDIS MAC address: %x-%x-%x-%x-%x-%x\r\n",(UCHAR) g_EbootCFG.RNDISMac[0], (UCHAR) (g_EbootCFG.RNDISMac[0]>>8), (UCHAR)g_EbootCFG.RNDISMac[1],
                                             (UCHAR)(g_EbootCFG.RNDISMac[1]>>8), (UCHAR)g_EbootCFG.RNDISMac[2], (UCHAR)(g_EbootCFG.RNDISMac[2]>>8));
                KITLOutputDebugString ( "6) %s image at startup\r\n", g_EbootCFG.autoDownloadImage?"Download new":"Launch existing flash resident");
                KITLOutputDebugString ( "7) Boot device order: ");
                {
                    if(g_EbootCFG.bootDeviceOrder >= NUM_BOOT_ORDERS)
                    g_EbootCFG.bootDeviceOrder = 0;
     
                    KITLOutputDebugString ( "%s\r\n", BootOrderString[g_EbootCFG.bootDeviceOrder]);
                }
                KITLOutputDebugString ( "8) Debug serial port: ");
                {
                    switch (g_EbootCFG.dwDbgSerPhysAddr)
                    {
                        case BULVERDE_BASE_REG_PA_FFUART:
                        KITLOutputDebugString ( "FFUART\r\n");
                        break;

                        case BULVERDE_BASE_REG_PA_BTUART:
                        default:
                        KITLOutputDebugString ( "BTUART\r\n");
                        break;
                    }
                }
                KITLOutputDebugString ( "D) Download image now\r\n");
                KITLOutputDebugString ( "L) Launch existing flash resident image now\r\n");
                KITLOutputDebugString ( "\r\n\r\nEnter your selection: ");
                Selection = 0;
                while (! ( ( (Selection >= '0') && (Selection <= '8') ) || 
                           ( (Selection == 'D') || (Selection == 'd') ) || 
                           ( (Selection == 'L') || (Selection == 'l') ) )) 
                {
                    Selection = OEMReadDebugByte();
                }
                KITLOutputDebugString ( "%c\r\n", Selection);
                switch (Selection)
                {
                case '0':
                    SetIP(&g_EbootCFG);
                    bCFGChanged=TRUE;
                    break;
                case '1':
                    SetMask(&g_EbootCFG);
                    bCFGChanged=TRUE;
                    break;
                case '2':
                    SetDelay(&g_EbootCFG);
                    bCFGChanged=TRUE;
                    break;
                case '3':
                    if (g_EbootCFG.DHCPEnable == TRUE)
                        g_EbootCFG.DHCPEnable = FALSE;
                    else
                        g_EbootCFG.DHCPEnable = TRUE;
                    bCFGChanged=TRUE;
                    break;
                case '4':
                    ResetDefaultEBootCFG(&g_EbootCFG);
                    bCFGChanged=TRUE;
                    break;
                case '5':
                    SetMAC(&rndis_mac[0]);
                    // Save the user enter RNDIS MAC address
                    g_EbootCFG.RNDISMac[0] = (rndis_mac[1] << 8) | rndis_mac[0];
                    g_EbootCFG.RNDISMac[1] = (rndis_mac[3] << 8) | rndis_mac[2];
                    g_EbootCFG.RNDISMac[2] = (rndis_mac[5] << 8) | rndis_mac[4];
                    bCFGChanged = TRUE;
                    break;
                case '6':
                    if (g_EbootCFG.autoDownloadImage == TRUE)
                        g_EbootCFG.autoDownloadImage = FALSE;
                    else
                        g_EbootCFG.autoDownloadImage = TRUE;
                    bCFGChanged=TRUE;
                    break;
                case '7':
                    g_EbootCFG.bootDeviceOrder++;
                    if (g_EbootCFG.bootDeviceOrder >= NUM_BOOT_ORDERS)
                        g_EbootCFG.bootDeviceOrder = 0;
                    bCFGChanged=TRUE;
                    break;
                case '8':
                    if ((g_EbootCFG.dwDbgSerPhysAddr == 0) || (g_EbootCFG.dwDbgSerPhysAddr == BULVERDE_BASE_REG_PA_BTUART))
                    {
                        g_EbootCFG.dwDbgSerPhysAddr = BULVERDE_BASE_REG_PA_FFUART;
                    }
                    else
                    {
                        g_EbootCFG.dwDbgSerPhysAddr = BULVERDE_BASE_REG_PA_BTUART;
                    }
                    bCFGChanged=TRUE;
                    break;
                case 'D':
                case 'd':
                    if (bCFGChanged == TRUE)
                    {
                        StoreEBootCFG(&g_EbootCFG);
                    }
                    g_DownloadImage = TRUE;
                    goto CONTINUE;
                case 'L':
                case 'l':
                    if (bCFGChanged == TRUE)
                    {
                        StoreEBootCFG(&g_EbootCFG);
                    }
                    g_DownloadImage = FALSE;
                    goto CONTINUE;

                default:
                    break;
                }
            }
        }
    }

  CONTINUE:

    // Provide the chosen debug serial port address to the OS image for debug output messages.
    //
    g_pBSPArgs->dbgSerPhysAddr = g_EbootCFG.dwDbgSerPhysAddr;

    BootDevice = -1;

	// Launching FLASH image  ...
	if ( !g_DownloadImage )
		return(TRUE);

    // Locate and initialize an Ethernet controller.
    //
    for(i = 0; (i < NUM_BOOT_DEVICES) && (BootDevice == -1) ; i++) 
    {
        // Try each of the boot devices in the specified order (until one succeeds or they all fail).
        //
        BootDevice = InitSpecifiedEthDevice(&g_pBSPArgs->kitl, BootOrder[g_EbootCFG.bootDeviceOrder][i]);
    }

    if( BootDevice == SDBG_USB_SERIAL )
    {
        // Indicate to OEMPreDownload that serial will be used
        g_SerialUSBDownload = TRUE;
        if(!USBSerialInit())
        {
            KITLOutputDebugString("ERROR: Failed to detect and initialize USB Function controller.\r\n");
            return(FALSE);
        }
    }

    // No device was found ... 
    //
    if (BootDevice == -1)
    {
        KITLOutputDebugString("ERROR: Failed to detect and initialize Ethernet controller.\r\n");
        return(FALSE);
    }
    else
    {
        // Make sure MAC address has been programmed.
        //
        if (!g_pBSPArgs->kitl.mac[0] && !g_pBSPArgs->kitl.mac[1] && !g_pBSPArgs->kitl.mac[2])
        {
           KITLOutputDebugString("ERROR: Invalid Ethernet address read from Ethernet controller.\n");
           return(FALSE);
        }
        else
        {
            KITLOutputDebugString("INFO: MAC address: %x-%x-%x-%x-%x-%x\r\n",
                  g_pBSPArgs->kitl.mac[0] & 0x00FF, g_pBSPArgs->kitl.mac[0] >> 8,
                  g_pBSPArgs->kitl.mac[1] & 0x00FF, g_pBSPArgs->kitl.mac[1] >> 8,
                  g_pBSPArgs->kitl.mac[2] & 0x00FF, g_pBSPArgs->kitl.mac[2] >> 8);
        }
    }

    return(TRUE);

}


//------------------------------------------------------------------------------
//
//  Function:  OEMPreDownload
//
//  Pre-download initialization routine.
//
DWORD OEMPreDownload(void)
{
    UINT32 SubnetMask;
    BOOL  fGotJumpImg = FALSE, fGotIP = FALSE;
    UINT32 DHCPLeaseTime = 0;
    UINT32 *pDHCPLeaseTime = &DHCPLeaseTime;
    UINT32 BootFlags = 0;
    const char * platformString = "MAINSTN"; 


    // Create device name based on Ethernet address (this is how Platform Builder identifies this device).
    //
    KITLOutputDebugString("+OEMPreDownload()\r\n");
    OALKitlCreateName(BSP_DEVICE_PREFIX, g_pBSPArgs->kitl.mac, g_pBSPArgs->deviceId);
    KITLOutputDebugString("INFO: Using device name: '%s'\n", g_pBSPArgs->deviceId);

    // User doesn't want to get a DHCP address...
    // If user wants to jump to existing image, we may still need to populate
    //  ARGS so that a KITL session can correctly start up.
    //
    if (g_EbootCFG.DHCPEnable == FALSE)
    {
        pDHCPLeaseTime             = NULL;
        g_pBSPArgs->kitl.ipAddress = g_EbootCFG.IP;
        g_pBSPArgs->kitl.ipMask    = g_EbootCFG.subnetMask;
        g_pBSPArgs->kitl.flags    &= ~OAL_KITL_FLAGS_DHCP;
    }

    // If user wants to jump to existing image - skip download...
    //
    if (!g_DownloadImage)
    {
        return(BL_JUMP);
    }

    if( g_SerialUSBDownload )
    {
        // Send boot requests indefinitely
        do
        {
            KITLOutputDebugString("Sending boot request...\r\n");
            if(!SerialSendBootRequest(platformString))
            {
                KITLOutputDebugString("Failed to send boot request\r\n");
                return BL_ERROR;
            }
        }
        while(!SerialWaitForBootAck(&fGotJumpImg));
        
        // Ack block zero to start the download
        SerialSendBlockAck(0);

        if( fGotJumpImg )
        {
            KITLOutputDebugString("Received boot request ack... jumping to image\r\n");
        }
        else
        {
            KITLOutputDebugString("Received boot request ack... starting download\r\n");
        }
        return fGotJumpImg  ? BL_JUMP : BL_DOWNLOAD;
    }

    // Initialize the TFTP transport.
    memcpy(g_DeviceAddr.wMAC, g_pBSPArgs->kitl.mac, (sizeof(UINT16) * 3));
    g_DeviceAddr.dwIP  = g_pBSPArgs->kitl.ipAddress;
    g_DeviceAddr.wPort = 0;
    SubnetMask         = g_pBSPArgs->kitl.ipMask;

    if (!EbootInitEtherTransport(&g_DeviceAddr,
                                 &SubnetMask,
                                 &fGotJumpImg,
                                 pDHCPLeaseTime,
                                 EBOOT_VERSION_MAJOR,
                                 EBOOT_VERSION_MINOR,
                                 BSP_DEVICE_PREFIX,
                                 g_pBSPArgs->deviceId,
                                 EDBG_CPU_ARM720,
                                 BootFlags))
    {
        return(BL_ERROR);
    }

    // If the user wanted a DHCP address, save the values obtained in the init call above.
    //
    if (g_EbootCFG.DHCPEnable == TRUE)
    {
        g_pBSPArgs->kitl.ipAddress  = g_DeviceAddr.dwIP;
        g_pBSPArgs->kitl.ipMask     = SubnetMask;
        g_pBSPArgs->kitl.flags     |= OAL_KITL_FLAGS_DHCP;
    }

    return(fGotJumpImg? BL_JUMP : BL_DOWNLOAD);
}


//------------------------------------------------------------------------------
//
//  Function:  OEMLaunch
//
//  Launch downloaded/stored image.
//
void OEMLaunch(DWORD dwImageStart, DWORD dwImageLength, DWORD dwLaunchAddr, const ROMHDR *pRomHdr)
{
    EDBG_OS_CONFIG_DATA *pCfgData;
    EDBG_ADDR EshellHostAddr;
    UINT32 PhysAddress;


    // If a launch address wasn't specified - use the last known good address.
    //
    if (!dwLaunchAddr)
    {
        dwLaunchAddr = g_EbootCFG.dwLaunchAddr;
    }
    else
    {
        if (g_EbootCFG.dwLaunchAddr != dwLaunchAddr ||
            g_EbootCFG.dwPhysStart  != dwImageStart ||
            g_EbootCFG.dwPhysLen    != dwImageLength)
        {
            g_EbootCFG.dwLaunchAddr = dwLaunchAddr;
            g_EbootCFG.dwPhysStart  = dwImageStart;
            g_EbootCFG.dwPhysLen    = dwImageLength;
            g_EbootCFG.dwStoreAddr  = 0;

            StoreEBootCFG(&g_EbootCFG);
        }
    }

    // Translate the image start address (virtual address) to a physical address.
    //
    PhysAddress = (UINT32) OALVAtoPA((VOID *)dwLaunchAddr);
    KITLOutputDebugString("Download successful!  Jumping to image at 0x%x (physical 0x%x)...\r\n", dwLaunchAddr, PhysAddress);

    // Wait for PB connection...
    //
    if (g_DownloadImage)
    {
        if( g_SerialUSBDownload )
        {
            // Wait indefinately for a jump request.
            // SerialWaitForJump() returns the debug transport requested by PB.
            // However, we only support ethernet for debug so we ignore the
            // return value.
            SerialWaitForJump();
        }
        else 
        {
            KITLOutputDebugString("Waiting for host to connect...\r\n");
            if (!(pCfgData = EbootWaitForHostConnect(&g_DeviceAddr, &EshellHostAddr)))
            {
                KITLOutputDebugString("ERROR: EbootWaitForHostConnect failed!\r\n");
                SpinForever();
            }

            // If the user selected "passive" KITL (i.e., don't connect to the target at boot time), set the
            // flag in the args structure so the OS image can honor it when it boots.
            //
            if (pCfgData && (pCfgData->KitlTransport & KTS_PASSIVE_MODE))
            {
                g_pBSPArgs->kitl.flags |= OAL_KITL_FLAGS_PASSIVE;
            }
        }
    }


    

    // Jump to the image we just downloaded.
    //
    KITLOutputDebugString("\r\n\r\n\r\n");
    Launch(PhysAddress);

    // Should never get here...
    //
    SpinForever();

}


//------------------------------------------------------------------------------
//
//  Function:  StoreEBootCFG
//
//  Stores bootloader configuration information (menu settings, etc.) in flash.
//
static BOOL StoreEBootCFG(EBOOT_CFG *EBootCFG)
{

    if (!FlashWrite(EBOOT_FLASH_CFG_START, (PUCHAR)EBootCFG, sizeof(EBOOT_CFG)))
    {
        KITLOutputDebugString("ERROR: StoreEBootCFG: failed to write configuration.\r\n");
        return(FALSE);
    }

    return(TRUE);
}


//------------------------------------------------------------------------------
//
//  Function:  ResetDefaultEBootCFG
//
//  Resets the debug bootloader configuration information (menu settings, etc.).
//
static void ResetDefaultEBootCFG(EBOOT_CFG *pEbootCFG)
{

    KITLOutputDebugString("\r\nResetting factory default configuration...\r\n");

    pEbootCFG->autoDownloadImage = TRUE;
    pEbootCFG->IP                = inet_addr("0.0.0.0"); 
    pEbootCFG->subnetMask        = inet_addr("0.0.0.0"); 
    pEbootCFG->numBootMe         = 50;
    pEbootCFG->delay             = 3;
    pEbootCFG->DHCPEnable        = TRUE;
    pEbootCFG->dwLaunchAddr      = (UINT32) OALPAtoVA(MAINSTONEII_BASE_PA_BOOT_FLASH, FALSE) + 0x80000;
    pEbootCFG->dwPhysStart       = 0;
    pEbootCFG->dwPhysLen         = 0;
    pEbootCFG->bootDeviceOrder   = 0;
    pEbootCFG->dwDbgSerPhysAddr  = BULVERDE_BASE_REG_PA_FFUART;
    pEbootCFG->dwStoreAddr       = 0;
    pEbootCFG->CheckSignatures   = FALSE;
    pEbootCFG->RNDISMac[0] = 0;
    pEbootCFG->RNDISMac[1] = 0;
    pEbootCFG->RNDISMac[2] = 0;
    pEbootCFG->ConfigMagicNumber = EBOOT_CFG_MAGIC_NUMBER;

}


//------------------------------------------------------------------------------
//
//  Function:  LoadEBootCFG
//
//  Retrieves bootloader configuration information (menu settings, etc.) from flash.
//
static BOOL LoadEBootCFG(EBOOT_CFG *EBootCFG)
{

    if (!FlashRead(EBOOT_FLASH_CFG_START, (PUCHAR)EBootCFG, sizeof(EBOOT_CFG)))
    {
        KITLOutputDebugString("ERROR: LoadEBootCFG: failed to load configuration.\r\n");
        return(FALSE);
    }

    // Is the CFG data valid?  Check for the magic number that was written the last time
    // the CFG block was updated.  If Eboot has never been run, there will be no configuration
    // information, so the magic number will likely not be found.  In this case, setup the 
    // factory defaults and store them into Flash.
    //
    if (EBootCFG->ConfigMagicNumber != EBOOT_CFG_MAGIC_NUMBER)
    {
        ResetDefaultEBootCFG(EBootCFG);
    }

    // Make sure a valid debug serial port address (physical) was found.
    //
    if (g_EbootCFG.dwDbgSerPhysAddr == 0)
    {
        g_EbootCFG.dwDbgSerPhysAddr = BULVERDE_BASE_REG_PA_BTUART;
    }

    return(TRUE);
}

//------------------------------------------------------------------------------
static void SetMAC(unsigned char macdata[6])
{
    BOOLEAN  bSuccess = TRUE;
    int idx;
    int macidx;
    unsigned char val;
    unsigned char cnt;
    char szDottedD[20]; // The string used to collect the dotted decimal IP address
    WORD cwNumChars = 0;
    UINT16 InChar = 0;


    KITLOutputDebugString ( "\r\nEnter RDNIS MAC address: ");

    while (!((InChar == 0x0d) || (InChar == 0x0a)))
    {
        InChar = OEMReadDebugByte();
        if (InChar != OEM_DEBUG_COM_ERROR && InChar != OEM_DEBUG_READ_NODATA)
        {
            // If it's a number or a period, add it to the string
            if (InChar == '-' || 
               (InChar >= '0' && InChar <= '9') || 
               (InChar >= 'a' && InChar <= 'f') || 
               (InChar >= 'A' && InChar <= 'F'))
            {
                if (cwNumChars < 20)
                {
                    szDottedD[cwNumChars++] = (char)InChar;
                    OEMWriteDebugByte((BYTE)InChar);
                }
            }
            // If it's a backspace, back up
            else if (InChar == 8)
            {
                if (cwNumChars > 0)
                {
                    cwNumChars--;
                    OEMWriteDebugByte((BYTE)InChar);
                }
            }
        }
    }

    // If it's a carriage return with an empty string, don't change anything.
    if (cwNumChars)
    {
        szDottedD[cwNumChars] = '-';
        szDottedD[cwNumChars+1] = '\0';

        idx = 0;
        macidx = 0;
        val = 0;
        cnt = 0;
        while ((szDottedD[idx] != 0) && (idx < 20))
        {
            if (szDottedD[idx] >= '0' && szDottedD[idx] <= '9') 
            {
               val <<= 4;
               val |= szDottedD[idx] - 0x30;
               cnt++;
            }
            else if (szDottedD[idx] >= 'a' && szDottedD[idx] <= 'f') 
            {
               val <<= 4;
               val |= szDottedD[idx] - 0x57;
               cnt++;
            }
            else if (szDottedD[idx] >= 'A' && szDottedD[idx] <= 'F')
            {
               val <<= 4;
               val |= szDottedD[idx] - 0x37;
               cnt++;
            }
            else
            {
               macdata[macidx] = val;
               macidx++;
               val = 0;
               cnt = 0;
            }
            idx++;

            if (cnt > 2)
            {
               KITLOutputDebugString("Invalid MAC Address format\r\n");
               bSuccess = FALSE;
               idx = 20;
            }
        } // endwhile
    }

    if (!bSuccess)
    {
       for (macidx = 0; macidx < 6; macidx++)
       {
          macdata[macidx] = 0;
       }
    }
}

static void SetIP(EBOOT_CFG *pEbootCFG)
{
    char szDottedD[16]; // The string used to collect the dotted decimal IP address
    WORD cwNumChars = 0;
    UINT16 InChar = 0;

    KITLOutputDebugString ( "\r\nEnter new IP address: ");

    while (!((InChar == 0x0d) || (InChar == 0x0a)))
    {
        InChar = OEMReadDebugByte();
        if (InChar != OEM_DEBUG_COM_ERROR && InChar != OEM_DEBUG_READ_NODATA)
        {
            // If it's a number or a period, add it to the string
            if (InChar == '.' || (InChar >= '0' && InChar <= '9'))
            {
                if (cwNumChars < 16)
                {
                    szDottedD[cwNumChars++] = (char)InChar;
                    OEMWriteDebugByte((BYTE)InChar);
                }
            }
            // If it's a backspace, back up
            else if (InChar == 8)
            {
                if (cwNumChars > 0)
                {
                    cwNumChars--;
                    OEMWriteDebugByte((BYTE)InChar);
                }
            }
        }
    }

    // If it's a carriage return with an empty string, don't change anything.
    if (cwNumChars)
    {
        szDottedD[cwNumChars] = '\0';
        pEbootCFG->IP = inet_addr( szDottedD );
    }
}


static void SetMask(EBOOT_CFG *pEbootCFG)
{
    char szDottedD[16]; // The string used to collect the dotted masks
    WORD cwNumChars = 0;
    UINT16 InChar = 0;

    KITLOutputDebugString ( "\r\nEnter new subnet mask: ");

    while (!((InChar == 0x0d) || (InChar == 0x0a)))
    {
        InChar = OEMReadDebugByte();
        if (InChar != OEM_DEBUG_COM_ERROR && InChar != OEM_DEBUG_READ_NODATA)
        {
            // If it's a number or a period, add it to the string
            if (InChar == '.' || (InChar >= '0' && InChar <= '9'))
            {
                if (cwNumChars < 16)
                {
                    szDottedD[cwNumChars++] = (char)InChar;
                    OEMWriteDebugByte((BYTE)InChar);
                }
            }
            // If it's a backspace, back up
            else if (InChar == 8)
            {
                if (cwNumChars > 0)
                {
                    cwNumChars--;
                    OEMWriteDebugByte((BYTE)InChar);
                }
            }
        }
    }

    // If it's a carriage return with an empty string, don't change anything.
    if (cwNumChars)
    {
        szDottedD[cwNumChars] = '\0';
        pEbootCFG->subnetMask = inet_addr( szDottedD );
    }
}


static void SetBootMe(EBOOT_CFG *pEbootCFG)
{
    char szCount[16];
    WORD cwNumChars = 0;
    UINT16 InChar = 0;

    KITLOutputDebugString ( "\r\nUse 0 for continuous boot me packets. \r\n");
    KITLOutputDebugString ( "Enter maximum number of boot me packets to send [0-255]: ");

    while (!((InChar == 0x0d) || (InChar == 0x0a)))
    {
        InChar = OEMReadDebugByte();
        if (InChar != OEM_DEBUG_COM_ERROR && InChar != OEM_DEBUG_READ_NODATA)
        {
            // If it's a number or a period, add it to the string
            if ((InChar >= '0' && InChar <= '9'))
            {
                if (cwNumChars < 16)
                {
                    szCount[cwNumChars++] = (char)InChar;
                    OEMWriteDebugByte((BYTE)InChar);
                }
            }
            // If it's a backspace, back up
            else if (InChar == 8)
            {
                if (cwNumChars > 0)
                {
                    cwNumChars--;
                    OEMWriteDebugByte((BYTE)InChar);
                }
            }
        }
    }

    // If it's a carriage return with an empty string, don't change anything.
    if (cwNumChars)
    {
        szCount[cwNumChars] = '\0';
        pEbootCFG->numBootMe = atoi(szCount);
        if (pEbootCFG->numBootMe > 255)
        {
            pEbootCFG->numBootMe = 255;
        }
        else if (pEbootCFG->numBootMe < 0)
        {
            pEbootCFG->numBootMe = 1;
        }
    }
}


static void SetDelay(EBOOT_CFG *pEbootCFG)
{
    char szCount[16];
    WORD cwNumChars = 0;
    UINT16 InChar = 0;

    KITLOutputDebugString ( "\r\nEnter maximum number of seconds to delay [1-255]: ");

    while (!((InChar == 0x0d) || (InChar == 0x0a)))
    {
        InChar = OEMReadDebugByte();
        if (InChar != OEM_DEBUG_COM_ERROR && InChar != OEM_DEBUG_READ_NODATA)
        {
            // If it's a number or a period, add it to the string
            if ((InChar >= '0' && InChar <= '9'))
            {
                if (cwNumChars < 16)
                {
                    szCount[cwNumChars++] = (char)InChar;
                    OEMWriteDebugByte((BYTE)InChar);
                }
            }
            // If it's a backspace, back up
            else if (InChar == 8)
            {
                if (cwNumChars > 0)
                {
                    cwNumChars--;
                    OEMWriteDebugByte((BYTE)InChar);
                }
            }
        }
    }

    // If it's a carriage return with an empty string, don't change anything.
    if (cwNumChars)
    {
        szCount[cwNumChars] = '\0';
        pEbootCFG->delay = atoi(szCount);
        if (pEbootCFG->delay > 255)
        {
            pEbootCFG->delay = 255;
        }
        else if (pEbootCFG->delay < 1)
        {
            pEbootCFG->delay = 1;
        }
    }
}


static ULONG bstrtoul(PUCHAR pStr, UCHAR nBase)
{
    UCHAR nPos=0;
    BYTE c;
    ULONG nVal = 0;
    UCHAR nCnt=0;
    ULONG n=0;

    // fulllibc doesn't implement isctype or iswctype, which are needed by
    // strtoul, rather than including coredll code, here's our own simple strtoul.

    if (pStr == NULL)
        return(0);

    for (nPos=0 ; nPos < strlen(pStr) ; nPos++)
    {
        c = tolower(*(pStr + strlen(pStr) - 1 - nPos));
        if (c >= '0' && c <= '9')
            c -= '0';
        else if (c >= 'a' && c <= 'f')
        {
            c -= 'a';
            c  = (0xa + c);
        }

        for (nCnt = 0, n = 1 ; nCnt < nPos ; nCnt++)
        {
            n *= nBase;
        }
        nVal += (n * c);
    }

    return(nVal);
}


BOOL OEMVerifyMemory(DWORD dwStartAddr, DWORD dwLength)
{
    BOOL  rc;
    DWORD dwPhysVerifyStart = (DWORD) OALVAtoPA((void *)dwStartAddr); 
    DWORD dwPhysVerifyEnd   = (DWORD) OALVAtoPA((void *)(dwStartAddr + dwLength - 1));


    KITLOutputDebugString("****** OEMVerifyMemory checking physical range [ 0x%x ==> 0x%x ]...\r\n", dwPhysVerifyStart, dwPhysVerifyEnd);


    if ((dwPhysVerifyStart >= IMAGE_BOOT_BLDRIMAGE_FLASH_PA_START) && (dwPhysVerifyEnd <= IMAGE_BOOT_BLDRIMAGE_FLASH_PA_END))
    {
        KITLOutputDebugString("****** bootloader address ****** \r\n\r\n");

        g_ImageType = IMAGE_TYPE_BOOTLOADER;
        rc = TRUE;
    }
    else if ((dwPhysVerifyStart >= MAINSTONEII_BASE_PA_BOOT_FLASH) && (dwPhysVerifyEnd <= (MAINSTONEII_BASE_PA_BOOT_FLASH + MAINSTONEII_SIZE_BOOT_FLASH)))
    {
        KITLOutputDebugString("****** FLASH address ****** \r\n\r\n");

        g_ImageType = IMAGE_TYPE_FLASHIMAGE;
        rc = TRUE;
    }
    else if ((dwPhysVerifyStart >= MAINSTONEII_BASE_PA_SDRAM) && (dwPhysVerifyEnd <= (MAINSTONEII_BASE_PA_SDRAM + MAINSTONEII_SIZE_SDRAM)))
    {
        KITLOutputDebugString("****** RAM address ****** \r\n\r\n");

        g_ImageType = IMAGE_TYPE_RAMIMAGE;
        rc = TRUE;
    }
    else
    {
        KITLOutputDebugString("****** OEMVerifyMemory FAILED - Invalid Memory Area ****** \r\n\r\n");

        rc = FALSE;
    }

    return(rc);

}

