//
// Copyright (c) Microsoft Corporation.  All rights reserved.
//
//
// Use of this sample source code is subject to the terms of the Microsoft
// license agreement under which you licensed this sample source code. If
// you did not accept the terms of the license agreement, you are not
// authorized to use this sample source code. For the terms of the license,
// please see the license agreement between you and Microsoft or, if applicable,
// see the LICENSE.RTF on your install media or the root of your tools installation.
// THE SAMPLE SOURCE CODE IS PROVIDED "AS IS", WITH NO WARRANTIES.
//
//
/**************************************************************************
** Copyright 2000-2004 Intel Corporation. All Rights Reserved.
**
** Portions of the source code contained or described herein and all documents
** related to such source code (Material) are owned by Intel Corporation
** or its suppliers or licensors and is licensed by Microsoft Corporation for distribution.  
** Title to the Material remains with Intel Corporation or its suppliers and licensors. 
** Use of the Materials is subject to the terms of the Microsoft license agreement which accompanied the Materials.  
** No other license under any patent, copyright, trade secret or other intellectual
** property right is granted to or conferred upon you by disclosure or
** delivery of the Materials, either expressly, by implication, inducement,
** estoppel or otherwise 
** Some portion of the Materials may be copyrighted by Microsoft Corporation.

Module Name:  

    CAMERA.c

**************************************************************************/
#include <windows.h>
#include <ceddk.h>
#include <types.h>
#include <string.h>
#include <stdio.h>
#include <tchar.h>
#include <nkintr.h>
#include <giisr.h>
#include <bulverde.h>

#include "xllp_ci.h"
#include "xllp_i2c.h"
#include "xllp_serialization.h"

#include "pxa27x_camera.h"


#define ZONE_ERROR 1
#define ZONE_INIT  1
#define ZONE_VERBOSE 1


#define DMA_INTERRUPT_REGISTER  (BULVERDE_BASE_REG_PA_DMAC + (DWORD)&(((P_XLLP_DMAC_T)0)->DINT))


// Global variables

// Pointer to camera driver instance which we will send back with callback functions
DWORD dwCameraDriverContext;    

// Signals the application that the video or still image frame is available
PFNCAMHANDLEFRAME pfnCameraHandleVideoFrame = NULL;
PFNCAMHANDLEFRAME pfnCameraHandleStillFrame = NULL;

volatile XLLP_DMAC_T   * v_pDMARegs = NULL;
volatile XLLP_GPIO_T   * v_pGPIORegs = NULL;
volatile XLLP_CLKMGR_T * v_pClkRegs = NULL;
volatile unsigned int  * v_pOSTRegs = NULL;
volatile unsigned int  * v_pCIRegs = NULL;
volatile unsigned int  * v_pI2C = NULL;


// Local variables
//
static const LPCWSTR g_wszDmaIsrDll     = L"giisr.dll";
static const LPCWSTR g_wszDmaIsrHandler = L"ISRHandler";

// The camera hardware context
static HW_CONTEXT_T g_HwContext;

// The MasterBufferList contains pointers to a DESCRIPTOR_INFO_T structure for each buffer in the list.
// This is currently statically allocated.
static DESCRIPTOR_INFO_T *MasterBufferList[MAX_CAMERA_DMA_BUFFERS]; 

// Describes the current number of buffers that are contained in the Master Buffer List.
static DWORD NumBuffers = 0;

// Address of the last phantom descriptor
static XLLP_VUINT32_T VirtAddr_Last_Pn_CH0 = 0x0;       
static XLLP_VUINT32_T VirtAddr_Last_Pn_CH1 = 0x0;
static XLLP_VUINT32_T VirtAddr_Last_Pn_CH2 = 0x0;

// Address of the last real descriptor
static XLLP_VUINT32_T VirtAddr_Last_n_CH0 = 0x0;        
static XLLP_VUINT32_T VirtAddr_Last_n_CH1 = 0x0;        
static XLLP_VUINT32_T VirtAddr_Last_n_CH2 = 0x0;        

// Virtual address of the phantom buffer
static XLLP_VUINT32_T VirtAddr_Phantom_Buffer = 0x0;
static XLLP_VUINT32_T VirtAddr_Phantom_Buffer_Base_Address = 0x0;

// Physical address of the phantom buffer
static XLLP_VUINT32_T PhysAddr_Phantom_Buffer = 0x0;

// Physical address of the start of the video capture descriptor chain
static XLLP_VUINT32_T VideoCaptureDescriptorPhysicalStartAddress_CH0 = 0x0;
static XLLP_VUINT32_T VideoCaptureDescriptorPhysicalStartAddress_CH1 = 0x0;
static XLLP_VUINT32_T VideoCaptureDescriptorPhysicalStartAddress_CH2 = 0x0;

// Physical address of the start of the still capture descriptor chain
static XLLP_VUINT32_T StillCaptureDescriptorPhysicalStartAddress_CH0 = 0x0;
static XLLP_VUINT32_T StillCaptureDescriptorPhysicalStartAddress_CH1 = 0x0;
static XLLP_VUINT32_T StillCaptureDescriptorPhysicalStartAddress_CH2 = 0x0;

// Set to true when the first video capture buffer has been submitted.
static BOOL bFirstVideoCaptureBufferSubmitted = FALSE;

// Set to true when the first stll image buffer has been submitted
static BOOL bFirstStillBufferSubmitted = FALSE;

// Used to provide synchronized access to certain camera functions
static CRITICAL_SECTION CameraMutex;

// Use to inform the interrupt service thread that a still capture event is occuring.
static unsigned int bDoStillCapture = FALSE;
static unsigned int bDoVideoCapture = FALSE;

// Used to discard a certain number of still image frames before signalling the application
// that the frame is available.  Some sensors output bad frames for the first few captured frames
// and they need to be "discarded."
static int discardframes = 0;

// Signals the application that a frame of motion video is available
static HANDLE hCameraFrameReady;

// Signals the application that the still image frame is available
static HANDLE hStillImageReady;

// Signals the application that the shutdown of the camera driver is complete.
static HANDLE hCameraShutdownComplete;

static HANDLE hShutDown;


static const XLLP_DMAC_DEVICE_T CAMERA_DMA_DEVICE[3] =
{
        XLLP_DMAC_QCI_RX1,
        XLLP_DMAC_QCI_RX2,
        XLLP_DMAC_QCI_RX3
};

static XLLP_STATUS_T PrvCameraDmacInit(HW_CONTEXT_T *pHwContext);
static XLLP_STATUS_T PrvCameraDmacDeinit(HW_CONTEXT_T *pHwContext);
static void PrvStartDMATransfer(P_HW_CONTEXT_T pHwContext, unsigned block_id);
static void PrvStopDMATransfer(P_HW_CONTEXT_T pHwContext);
static XLLP_STATUS_T PrvStartCapture(P_HW_CONTEXT_T pHwContext, unsigned int block_id, unsigned int frames);
static XLLP_STATUS_T PrvCameraSetCaptureFormat(P_HW_CONTEXT_T pHwContext);

static DWORD WINAPI CAMIntrThread(HW_CONTEXT_T *pHwContext);

VOID CameraInterfaceInterruptHandler(HW_CONTEXT_T *pHwContext)
{
    DWORD dwDCSR;
    DWORD dwDINT;
    int i;

    dwDINT = v_pDMARegs->DINT;
    DEBUGMSG(ZONE_VERBOSE, (TEXT( "CameraInterfaceInterruptHandler: DINT=0x%x\r\n" ), dwDINT));

    for(i=0; i<3; i++)
    if(dwDINT & (1<<pHwContext->XllpDMA.dma_channels[i]))
    {
        dwDCSR  = v_pDMARegs->DCSR[pHwContext->XllpDMA.dma_channels[i]];
        DEBUGMSG(ZONE_VERBOSE, (TEXT( "CameraInterfaceInterruptHandler: CH[%d] DCSR: 0x%x\r\n" ),pHwContext->XllpDMA.dma_channels[i], dwDCSR));

        // Provide waiting Object the current value of DCSR
        // SetEventData(pHwContext->hDMAEvent[i], dwDCSR);

        //did we get an error?
        if (dwDCSR & XLLP_DMAC_DCSR_BUS_ERR_INTR)
        {
            DEBUGMSG(ZONE_ERROR, (TEXT( "DCSR_BUSERRINTR ERROR\r\n" )) );
            DEBUGCHK(0);  // an unplanned dma interrupt occured 
        }
    
        //did we get an start interrupt?
        if (dwDCSR & XLLP_DMAC_DCSR_START_INTR)  
        {
            DEBUGMSG(ZONE_VERBOSE, (TEXT( "DCSR_STARTINTR\n" )) );
            DEBUGCHK(0); // an unplanned dma interrupt occured 
        }
    
        //did we get an end interrupt?
        if (dwDCSR & XLLP_DMAC_DCSR_END_INTR)  //if the channel stopped at the end of a buffer xfer
        {
            DEBUGMSG(ZONE_ERROR, (TEXT( "DCSR_ENDINTR\r\n" )) );
        }

        //did we get an stop interrupt?
        if (dwDCSR & XLLP_DMAC_DCSR_STOP_INTR)  
        {
            DEBUGMSG(ZONE_VERBOSE, (TEXT( "DCSR_STOPINTR\r\n" )) );
            dwDCSR &= ~XLLP_DMAC_DCSR_STOP_IRQ_EN;
            DEBUGCHK(0); // an unplanned dma interrupt occured 
        }

        // Clear the status
        v_pDMARegs->DCSR[pHwContext->XllpDMA.dma_channels[i]] = dwDCSR;

        SetEvent(pHwContext->hDMAEvent[i]);
    }

    return;
}


DWORD DMAInterruptThread(HW_CONTEXT_T *pHwContext)
{
    DEBUGMSG(1, (TEXT("DMAInterruptThread\r\n")));

    while (TRUE)
    {
        WaitForSingleObject(pHwContext->hDmaIntrEvent,  INFINITE);
        // Grab the lock
        //Lock();

        CameraInterfaceInterruptHandler(pHwContext);

        //Unlock();
        
        InterruptDone(pHwContext->dwSysIntrDMA);

    }  // while(TRUE)

  return 0;
}

XLLP_STATUS_T   PrvCameraDmacInit(HW_CONTEXT_T *pHwContext)
{
    int i;
    _TCHAR EventName[20];
    DWORD dwDMAIrq = pHwContext->dwDmaIrq;

    for(i=0; i<3; i++)
    {
        // Get DMA channel event handler
        // Compose the name of the channel event for use with CreateEvent(...) below.
        wsprintf(EventName, TEXT("DMA_CHANNEL_%d"), pHwContext->XllpDMA.dma_channels[i]);

        pHwContext->hDMAEvent[i] = CreateEvent(NULL,FALSE,FALSE, EventName);
        if (pHwContext->hDMAEvent[i] == NULL)
        {
            DEBUGMSG(ZONE_ERROR,(TEXT("hDMAEvent (Named) is Zero...\r\n")));
            return (FALSE);  
        }
    }

    // Convert the hardwre DMA controller interrupt IRQ into a logical SYSINTR value
    if(!pHwContext->dwSysIntrDMA || pHwContext->dwSysIntrDMA == SYSINTR_UNDEFINED)
    {
        if(!KernelIoControl(IOCTL_HAL_REQUEST_SYSINTR,
                            &dwDMAIrq, sizeof(DWORD),
                            &pHwContext->dwSysIntrDMA, sizeof(DWORD), NULL))
        {
            DEBUGMSG(ZONE_ERROR, (TEXT("Error obtaining DMA interrupt SYSINTR value!\n")));
            return (FALSE);
        }
    }

    // Install the DMA ISR handler
    if(pHwContext->hDMAIsrHandler == NULL)
    {
        GIISR_INFO Info;
        HANDLE hBusAccessHandle;
        PVOID PhysAddr;
        DWORD inIoSpace = 0;    // io space
        PHYSICAL_ADDRESS DmaRegisterAddress = {DMA_INTERRUPT_REGISTER, 0}; 

        pHwContext->hDMAIsrHandler = LoadIntChainHandler(g_wszDmaIsrDll, 
                                               g_wszDmaIsrHandler,
                                               (BYTE)dwDMAIrq);
        if(pHwContext->hDMAIsrHandler == NULL)
        {
            DEBUGMSG(ZONE_ERROR, (TEXT("LoadIntChainHandler (%s, %s, %d) failed!\r\n"),
                                            g_wszDmaIsrDll, 
                                            g_wszDmaIsrHandler,
                                            dwDMAIrq));
            return FALSE;
        }

        hBusAccessHandle = CreateBusAccessHandle((LPCTSTR)pHwContext->hDriverIndex);
        if(!BusTransBusAddrToStatic(hBusAccessHandle,
                                    Internal, 0,
                                    DmaRegisterAddress, sizeof(DWORD),
                                    &inIoSpace, &PhysAddr))
        {
            DEBUGMSG(ZONE_ERROR, (L"pxa27x_camera: Failed TransBusAddrToStatic\r\n"));
            return FALSE;
        }
            
        DEBUGMSG(ZONE_INIT, (L"Installed ISR handler, Dll = '%s', Handler = '%s', Irq = %d, PhysAddr = 0x%x\r\n", 
                g_wszDmaIsrDll, g_wszDmaIsrHandler, dwDMAIrq, PhysAddr));

        // Set up ISR handler
        Info.SysIntr    = pHwContext->dwSysIntrDMA;
        Info.CheckPort  = TRUE;
        Info.PortIsIO   = FALSE;
        Info.UseMaskReg = FALSE;
        Info.PortAddr   = (DWORD)PhysAddr;
        Info.PortSize   = sizeof(DWORD);
        Info.Mask       = (1 << pHwContext->XllpDMA.dma_channels[0]) |
                          (1 << pHwContext->XllpDMA.dma_channels[1]) |
                          (1 << pHwContext->XllpDMA.dma_channels[2]) ;
            
        if (!KernelLibIoControl(pHwContext->hDMAIsrHandler, IOCTL_GIISR_INFO, &Info, sizeof(Info), NULL, 0, NULL))
        {
            DEBUGMSG(ZONE_ERROR, (L"KernelLibIoControl call failed.\r\n"));
        }

        CloseBusAccessHandle(hBusAccessHandle);
    }

    //Create event for dwSysIntrDMA
    pHwContext->hDmaIntrEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
    if (!pHwContext->hDmaIntrEvent)
    {
        return FALSE; 
    }
    if (!InterruptInitialize (pHwContext->dwSysIntrDMA, pHwContext->hDmaIntrEvent, NULL, 0))
    {
        return FALSE;
    }

    //create and start hDmaIntrThread
    pHwContext->hDmaIntrThread = CreateThread((LPSECURITY_ATTRIBUTES)NULL,
                                            0,
                                            (LPTHREAD_START_ROUTINE)DMAInterruptThread,
                                            (void*)pHwContext,
                                            0,
                                            NULL);
    if (!pHwContext->hDmaIntrThread)
    {
        return FALSE;
    }

    //CeSetThreadPriority(pHwContext->hDmaIntrThread, GetInterruptThreadPriority());

    return XLLP_STATUS_SUCCESS;
}


XLLP_STATUS_T   PrvCameraDmacDeinit(HW_CONTEXT_T *pHwContext)
{
    return XLLP_STATUS_SUCCESS;
}

int CameraInit(void *pData)
{
    PHYSICAL_ADDRESS PA;
    HANDLE hThread;

    PA.QuadPart  = BULVERDE_BASE_REG_PA_DMAC;
    v_pDMARegs   = (volatile XLLP_DMAC_T *) MmMapIoSpace(PA, sizeof(XLLP_DMAC_T), FALSE);

    PA.QuadPart  = BULVERDE_BASE_REG_PA_I2C;
    v_pI2C  = (volatile unsigned int *) MmMapIoSpace(PA, sizeof(XLLP_I2C_T), FALSE);
    if (!v_pI2C)
    {
        return FALSE;
    }

    PA.QuadPart  = BULVERDE_BASE_REG_PA_OST;
    v_pOSTRegs  = (volatile unsigned int *) MmMapIoSpace(PA, sizeof(XLLP_OST_T), FALSE);
    if (!v_pOSTRegs)
    {
        return FALSE;
    }

    PA.QuadPart  = BULVERDE_BASE_REG_PA_CAMERA;
    v_pCIRegs  = (volatile unsigned int *) MmMapIoSpace(PA, 1024, FALSE);
    if (!v_pCIRegs)
    {
        return FALSE;
    }
        
    PA.QuadPart  = BULVERDE_BASE_REG_PA_CLKMGR;
    v_pClkRegs  = (volatile XLLP_CLKMGR_T *) MmMapIoSpace(PA, sizeof(XLLP_CLKMGR_T), FALSE);
    if (!v_pClkRegs)
    {
        return FALSE;
    }

    PA.QuadPart  = BULVERDE_BASE_REG_PA_GPIO;
    v_pGPIORegs  = (volatile XLLP_GPIO_T *) MmMapIoSpace(PA, sizeof(XLLP_GPIO_T), FALSE);
    if (!v_pGPIORegs)
    {
        return FALSE;
    }

    //OS_I2CInit();
    XllpI2cInit((P_XLLP_I2C_T)(v_pI2C), (P_XLLP_GPIO_T) v_pGPIORegs, (P_XLLP_CLKMGR_T) v_pClkRegs, (XLLP_UINT32_T) 0);

    VirtAddr_Last_Pn_CH0 = 0x0;     
    VirtAddr_Last_Pn_CH1 = 0x0;
    VirtAddr_Last_Pn_CH2 = 0x0;
    VirtAddr_Last_n_CH0 = 0x0;      
    VirtAddr_Last_n_CH1 = 0x0;      
    VirtAddr_Last_n_CH2 = 0x0;      
    VirtAddr_Phantom_Buffer = 0x0;
    VirtAddr_Phantom_Buffer_Base_Address = 0x0;
    PhysAddr_Phantom_Buffer = 0x0;
    bFirstVideoCaptureBufferSubmitted = FALSE;
    bFirstStillBufferSubmitted = FALSE;
    NumBuffers = 0;

    hCameraFrameReady = CreateEvent(NULL,FALSE,FALSE,CAMERA_EVENT_NAME);
    hStillImageReady = CreateEvent(NULL, FALSE, FALSE, CAMERA_EVENT_STILL_IMAGE_READY);
    hCameraShutdownComplete = CreateEvent(NULL, FALSE, FALSE, CAMERA_SHUTDOWN_COMPLETE);
    hShutDown = CreateEvent(NULL, FALSE, FALSE, NULL);

    memset(&g_HwContext.XllpCAMERA, 0, sizeof(XLLP_Camera_Context_T));
    memset(&g_HwContext.XllpSENSOR, 0, sizeof(XLLP_Camera_Function_T));
    memset(&g_HwContext.XllpDMA, 0, sizeof(XLLP_Camera_DMA_Context_T));

    // Fill in the context structure
    g_HwContext.hDriverIndex = pData;
    g_HwContext.dwDmaIrq = IRQ_DMAC;

    g_HwContext.XllpCAMERA.camera_functions = &g_HwContext.XllpSENSOR;

    g_HwContext.XllpCAMERA.capture_mode = XLLP_CAMERA_MODE_RECORD;

    // os mapped register address   
    g_HwContext.XllpCAMERA.clk_reg_base  = (unsigned int)v_pClkRegs;
    g_HwContext.XllpCAMERA.ost_reg_base  = (unsigned int)v_pOSTRegs;
    g_HwContext.XllpCAMERA.gpio_reg_base = (unsigned int)v_pGPIORegs;
    g_HwContext.XllpCAMERA.ci_reg_base   = (unsigned int)v_pCIRegs;
    InitializeCriticalSection(&CameraMutex);

    XllpDmacInitHandle((P_XLLP_DMAC_T)v_pDMARegs);

    if(!CameraInitSensor())
    {
        DEBUGMSG( ZONE_ERROR, ( _T("pxa27x_camera: CameraInitSensor() failed\r\n") ) );
        return FALSE;
    }

    PrvCameraDmacInit(&g_HwContext);

    // Launch the Camera ISR
    hThread = CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)CAMIntrThread,(void*)&g_HwContext,0,NULL);
    CloseHandle(hThread);

    // return camera context
    return (int)&g_HwContext.XllpCAMERA;
} 

int  CameraDeinit(void)
{
    RETAILMSG(1, (TEXT("CameraDeinit\r\n")));

    SensorDeinitPlatform(&g_HwContext.XllpCAMERA, &g_HwContext.XllpDMA);

    //OS_I2CDeinit();

    if (v_pDMARegs)
    {
        MmUnmapIoSpace((void *)v_pDMARegs, sizeof(XLLP_DMAC_T));
        v_pDMARegs = NULL;
    }

    if (v_pI2C)
    {
        MmUnmapIoSpace((PVOID)v_pI2C, sizeof(XLLP_I2C_T));
        v_pI2C = NULL;
    }

    if (v_pOSTRegs)
    {
        MmUnmapIoSpace((PVOID)v_pOSTRegs, sizeof(XLLP_OST_T));
        v_pOSTRegs = NULL;
    }

    if (v_pCIRegs)
    {
        MmUnmapIoSpace((PVOID)v_pCIRegs, 1024);
        v_pCIRegs = NULL;
    }
        
    if (v_pClkRegs)
    {
        MmUnmapIoSpace((PVOID)v_pClkRegs, sizeof(XLLP_CLKMGR_T));
        v_pClkRegs = NULL;
    }

    if (v_pGPIORegs)
    {
        MmUnmapIoSpace((PVOID)v_pGPIORegs, sizeof(XLLP_GPIO_T));
        v_pGPIORegs = NULL;
    }
    DeleteCriticalSection(&CameraMutex);

    CloseHandle(hCameraFrameReady);
    CloseHandle(hStillImageReady);
    CloseHandle(hCameraShutdownComplete);
    CloseHandle(hShutDown);

    return TRUE;
}


// The interrupt service thread for the camera driver.
// When an end of frame interrupt occurs, this thread signals the application through an event mechanism.
// There's one event to signal a video frame has been captured, and another event to signal that a still image buffer
// frame has been captured.

DWORD WINAPI CAMIntrThread(HW_CONTEXT_T *pHwContext)
{
    HANDLE events[2];
    DWORD event_id = 0;

    CeSetThreadPriority(GetCurrentThread(), 1);

    events[0] = pHwContext->hDMAEvent[0];
    events[1] = hShutDown;

    while((event_id - WAIT_OBJECT_0) != 1) 
    {
        event_id = WaitForMultipleObjects(2, (CONST HANDLE*)events, FALSE, INFINITE);

        DEBUGMSG(1, (TEXT("CAMIntrThread: got event\r\n")));

        // Service the "DMA transfer complete" interrupt
        // which means that a new frame of video is available.
        if ((event_id - WAIT_OBJECT_0) == 0)
        {
            CameraInterruptHandler(&g_HwContext);

            // When we perform a still capture, discard n frames of video before we signal the application
            // that the data is ready
            if ((!bDoStillCapture)&& bDoVideoCapture)
            {
                SetEvent(hCameraFrameReady);            

                if (pfnCameraHandleVideoFrame)
                    pfnCameraHandleVideoFrame( dwCameraDriverContext );
            }
            
            if (discardframes && bDoStillCapture)
            {
                discardframes--;
            }

            if (bDoStillCapture && (discardframes == 0))
            {
                SetEvent(hStillImageReady);
                ResetEvent(hCameraFrameReady);
                bDoStillCapture = FALSE;

                if (pfnCameraHandleStillFrame)
                     pfnCameraHandleStillFrame( dwCameraDriverContext );
            }
        }
    }

    SetEvent(hCameraShutdownComplete);
    ExitThread(0);
    return TRUE;
} 


void CameraInterruptHandler(HW_CONTEXT_T *pHwContext)
{
    DEBUGMSG(1, (TEXT("CameraInterruptHandler\r\n")));
}


// Initialize the XllpCAMERA structure with driver and application specific parameters.
int CameraInitSensor(void)
{
    XLLP_STATUS_T status;

    // platform specific initialization
    if(!SensorInitPlatform(&g_HwContext.XllpCAMERA, &g_HwContext.XllpDMA))
    {
        return FALSE;
    }

    XllpLock(I2C);

    status = XllpCameraInit(&g_HwContext.XllpCAMERA);
    if (status != XLLP_STATUS_SUCCESS)
    {
            RETAILMSG(1,(TEXT("Camera init failure!!!\r\n")));
            return FALSE;
    }

    XllpUnlock(I2C);

    return TRUE;
}

void CameraWriteRegister(unsigned char subAddress, unsigned char *bufP)
{
        
    if(g_HwContext.XllpCAMERA.camera_functions->WriteRegister)
    {
        XllpLock(I2C);
        g_HwContext.XllpCAMERA.camera_functions->WriteRegister(&g_HwContext.XllpCAMERA, subAddress, bufP);
        CameraReadRegister(subAddress, bufP);
        XllpUnlock(I2C);
    }
}

void CameraReadRegister(unsigned char subAddress, unsigned char *bufP)
{
    *bufP = 0;

    if(g_HwContext.XllpCAMERA.camera_functions->ReadRegister)
    {
        XllpLock(I2C);
        g_HwContext.XllpCAMERA.camera_functions->ReadRegister(&g_HwContext.XllpCAMERA, subAddress, bufP);
        XllpUnlock(I2C);
    }
}

// Start the video capture process initiated by the application
void CameraStartVideoCapture()
{
    XLLP_STATUS_T status;

    EnterCriticalSection(&CameraMutex);

    bDoStillCapture = FALSE;
    bDoVideoCapture = TRUE;

    XllpLock(I2C);

    // Configure the capture format.  Set the width, height, mode, and the physical addresses of the 
    // start of the descriptor chain.
    g_HwContext.XllpCAMERA.capture_width = g_HwContext.XllpCAMERA.Video_capture_width;
    g_HwContext.XllpCAMERA.capture_height = g_HwContext.XllpCAMERA.Video_capture_height;
    g_HwContext.XllpCAMERA.capture_mode = XLLP_CAMERA_MODE_RECORD;
    g_HwContext.XllpCAMERA.sensor_flip_mode = XLLP_CAMERA_VIDEO_FLIP_VERTICAL|
                                              XLLP_CAMERA_VIDEO_FLIP_HORIZONTAL;

    g_HwContext.XllpDMA.fifo0_descriptors_physical = VideoCaptureDescriptorPhysicalStartAddress_CH0;
    g_HwContext.XllpDMA.fifo1_descriptors_physical = VideoCaptureDescriptorPhysicalStartAddress_CH1;
    g_HwContext.XllpDMA.fifo2_descriptors_physical = VideoCaptureDescriptorPhysicalStartAddress_CH2;

    // set capture format
    status = PrvCameraSetCaptureFormat(&g_HwContext);
    if (status)
    {
        RETAILMSG(1,(TEXT("Camera init failure inside CameraSetCaptureFormat().\r\n")));
    }

    XllpCameraSetInterruptMask(&g_HwContext.XllpCAMERA, 
                               XLLP_CI_CICR0_FOM            | 
                               XLLP_CI_CICR0_EOFM           | 
                               XLLP_CI_CICR0_SOFM           | 
                               XLLP_CI_CICR0_CDM            |
                               XLLP_CI_CICR0_QDM            |
                               XLLP_CI_CICR0_PERRM          |
                               XLLP_CI_CICR0_EOLM           | 
                               XLLP_CI_CICR0_FEM);

    PrvStartCapture(&g_HwContext, 0, 0);
    
    XllpUnlock(I2C);

    LeaveCriticalSection(&CameraMutex);
}

void CameraStopVideoCapture()
{
    EnterCriticalSection(&CameraMutex);
    bDoVideoCapture = FALSE;

    XllpLock(I2C);

    // stop capture
    g_HwContext.XllpCAMERA.camera_functions->stop_capture(&g_HwContext.XllpCAMERA);    

    // stop dma
    PrvStopDMATransfer(&g_HwContext);

    XllpUnlock(I2C);

    LeaveCriticalSection(&CameraMutex);
}

void CameraCaptureStillImage()
{
    XLLP_STATUS_T status;
    EnterCriticalSection(&CameraMutex);

    XllpLock(I2C);

    // stop capture
    g_HwContext.XllpCAMERA.camera_functions->stop_capture(&g_HwContext.XllpCAMERA);    

    // stop dma
    PrvStopDMATransfer(&g_HwContext);



    discardframes = 0;

    // Configure the capture format.  Set the width, height, mode, and the physical addresses of the 
    // start of the descriptor chain.
    g_HwContext.XllpCAMERA.capture_width  = g_HwContext.XllpCAMERA.Still_capture_width;
    g_HwContext.XllpCAMERA.capture_height = g_HwContext.XllpCAMERA.Still_capture_height;
    g_HwContext.XllpCAMERA.capture_mode   = XLLP_CAMERA_MODE_STILL;
    g_HwContext.XllpCAMERA.sensor_flip_mode = XLLP_CAMERA_VIDEO_FLIP_HORIZONTAL |
                                              XLLP_CAMERA_STILL_FLIP_HORIZONTAL;

    g_HwContext.XllpDMA.fifo0_descriptors_physical = StillCaptureDescriptorPhysicalStartAddress_CH0;
    g_HwContext.XllpDMA.fifo1_descriptors_physical = StillCaptureDescriptorPhysicalStartAddress_CH1;
    g_HwContext.XllpDMA.fifo2_descriptors_physical = StillCaptureDescriptorPhysicalStartAddress_CH2;

    // Reset the buffer chain.  The application must now resubmit all capture buffers.
    bFirstVideoCaptureBufferSubmitted = FALSE;
    VirtAddr_Last_Pn_CH0 = 0x0;
    
    // set capture format
    status = PrvCameraSetCaptureFormat(&g_HwContext);
    if (status)
    {
            RETAILMSG(1,(TEXT("Camera init failure inside CameraSetCaptureFormat().\r\n")));
    //      return 1;
    }

    XllpCameraSetInterruptMask(&g_HwContext.XllpCAMERA, 
                               XLLP_CI_CICR0_FOM            | 
                               XLLP_CI_CICR0_EOFM           | 
                               XLLP_CI_CICR0_SOFM           | 
                               XLLP_CI_CICR0_CDM            |
                               XLLP_CI_CICR0_QDM            |
                               XLLP_CI_CICR0_PERRM          |
                               XLLP_CI_CICR0_EOLM           | 
                               XLLP_CI_CICR0_FEM);
    
    bDoStillCapture = TRUE;
    bDoVideoCapture = FALSE;

    // Enable the still image capture sequence.
    // Calling this routine once results in capture sequence that sometimes fails to interrupt
    // on the end of frame.
    // Calling this routine twice works around this problem - but it needs to be looked into
    // to root cause the failure.
 //   PrvStartCapture(&g_HwContext, 0, 1);
    PrvStartCapture(&g_HwContext, 0, 1);

    XllpUnlock(I2C);

    LeaveCriticalSection(&CameraMutex);
}

// Shut down this session of the camera driver and free all of the relevent resources allocated within.
int CameraShutdown(void)
{
   
    XllpLock(I2C);

    XllpCameraDeInit(&g_HwContext.XllpCAMERA);

    // Signal the ISR thread to shutdown
    SetEvent(hShutDown);

    XllpUnlock(I2C);

    return TRUE;
}

// A helper function that is called to configure the parameters that are later used 
// to configure the capture format.
int CameraChangeCaptureFormat(void)
{
    // TBD: call some real function
    return TRUE;
}


// Provides a physical address for a given user mode virtual address.
BOOL getPhyAddress (DWORD * pdwVirtual, DWORD * pdwPhysical)
{
    BOOL bReturnVal = TRUE;
    DWORD dwPageSize = UserKInfo[KINX_PAGESIZE];
    DWORD dwPageTableEntry;
    DWORD dwPageTablePart;
    DWORD dwOffsetInPage;

    /* query the page tables */
    if (LockPages (pdwVirtual, sizeof (DWORD), &dwPageTableEntry, LOCKFLAG_QUERY_ONLY) != TRUE)
    {
            bReturnVal = FALSE;
            goto cleanAndReturn;
    }

    /* shift left to make into a real address */
    dwPageTablePart = dwPageTableEntry << UserKInfo[KINX_PFN_SHIFT];

    /* figure out the offset in the page */
    dwOffsetInPage = ((DWORD) pdwVirtual) & (~UserKInfo[KINX_PFN_MASK]);

    /* combine the page table part and offset to create the address */
    *pdwPhysical = dwPageTablePart | dwOffsetInPage;

    cleanAndReturn:

    return (bReturnVal);
}
void CameraUnprepareBuffer(P_CAMERA_DMA_BUFFER_INFO pCamDmaBuf, int ucBufferType)
{
    DESCRIPTOR_INFO_T *pDescriptorInfo;

    pDescriptorInfo = NULL;
    bDoStillCapture = FALSE;
    bDoVideoCapture = FALSE;
    if (pCamDmaBuf)
    {
        if (pCamDmaBuf->BufferID >=0 && pCamDmaBuf->BufferID < MAX_CAMERA_DMA_BUFFERS)
        {
            pDescriptorInfo = MasterBufferList[pCamDmaBuf->BufferID];
        }
    }
    if (pDescriptorInfo)
    {
        free(pDescriptorInfo->pDescriptorBufferPhyAddress);
        free(pDescriptorInfo->pDescriptorBufferVirtAddress);
        free(pDescriptorInfo->pBufferPhyAddress);
        free(pDescriptorInfo);
        MasterBufferList[pCamDmaBuf->BufferID]= NULL;        
    }
    if (NumBuffers > 0)
    {
        NumBuffers--;
    }
        // We are going to unprepare all Video or still buffers
        // So we should initialize all related variables also

    if (ucBufferType == VIDEO_CAPTURE_BUFFER)
    {
        VirtAddr_Last_Pn_CH0 = 0x0;       
        VirtAddr_Last_Pn_CH1 = 0x0;
        VirtAddr_Last_Pn_CH2 = 0x0;

        // Address of the last real descriptor
        VirtAddr_Last_n_CH0 = 0x0;        
        VirtAddr_Last_n_CH1 = 0x0;        
        VirtAddr_Last_n_CH2 = 0x0;  

        // Physical address of the start of the video capture descriptor chain
        VideoCaptureDescriptorPhysicalStartAddress_CH0 = 0x0;
        VideoCaptureDescriptorPhysicalStartAddress_CH1 = 0x0;
        VideoCaptureDescriptorPhysicalStartAddress_CH2 = 0x0;

         // Set to true when the first video capture buffer has been submitted.
        bFirstVideoCaptureBufferSubmitted = FALSE;
    }
    else
    {
        // Physical address of the start of the still capture descriptor chain
        StillCaptureDescriptorPhysicalStartAddress_CH0 = 0x0;
        StillCaptureDescriptorPhysicalStartAddress_CH1 = 0x0;
        StillCaptureDescriptorPhysicalStartAddress_CH2 = 0x0;


        // Set to true when the first stll image buffer has been submitted
        bFirstStillBufferSubmitted = FALSE;


    }
    if (NumBuffers == 0)
    {
        if (VirtAddr_Phantom_Buffer_Base_Address)
        {
            free((void *)VirtAddr_Phantom_Buffer_Base_Address);
        }

        // Virtual address of the phantom buffer
        VirtAddr_Phantom_Buffer = 0x0;
        VirtAddr_Phantom_Buffer_Base_Address = 0x0;

        // Physical address of the phantom buffer
        PhysAddr_Phantom_Buffer = 0x0;
    }

    return;
}



//
// Submit a buffer to be filled by the camera DMA.
// A buffer is submitted by chaining the last buffer to this new buffer.
//
// Requires a valid FrameID, which is obtained from a successful call to PrepareBuffer().
//
void CameraSubmitBuffer(P_CAMERA_DMA_BUFFER_INFO pBufInfo, int BufferType)
{
    DESCRIPTOR_INFO_T *pDescriptorInfo;
    XLLP_DMAC_DESCRIPTOR_T *pXPn_CH0;
    XLLP_DMAC_DESCRIPTOR_T *pXPn_CH1;
    XLLP_DMAC_DESCRIPTOR_T *pXPn_CH2;

    XLLP_DMAC_DESCRIPTOR_T *pXn_CH0;
    XLLP_DMAC_DESCRIPTOR_T *pXn_CH1;
    XLLP_DMAC_DESCRIPTOR_T *pXn_CH2;

    XLLP_DMAC_DESCRIPTOR_T *pLast_Pn_CH0;
    XLLP_DMAC_DESCRIPTOR_T *pLast_Pn_CH1;
    XLLP_DMAC_DESCRIPTOR_T *pLast_Pn_CH2;

    XLLP_DMAC_DESCRIPTOR_T *pLast_n_CH0;
    XLLP_DMAC_DESCRIPTOR_T *pLast_n_CH1;
    XLLP_DMAC_DESCRIPTOR_T *pLast_n_CH2;

    BOOL isYUV_PLANAR;

    pDescriptorInfo = NULL;
    if (pBufInfo)
    {
        if (pBufInfo->BufferID >=0 && pBufInfo->BufferID < MAX_CAMERA_DMA_BUFFERS)
        {
                pDescriptorInfo = MasterBufferList[pBufInfo->BufferID];
        }
    }
    if (pDescriptorInfo)
    {
        isYUV_PLANAR = (3 == pDescriptorInfo->dwPlanesNumber ? TRUE : FALSE);

        // If this is the first buffer being submitted, then we want to store the physical address of the descriptor chain
        // so that when we start DMA's later on, we start with _this_ buffer.  All buffers submitted after this first buffer
        // are chained behind this one.  Therefore, this buffer is the logical choice to start with.    
        if (BufferType == VIDEO_CAPTURE_BUFFER)
        {
            if (bFirstVideoCaptureBufferSubmitted == FALSE)
            {
                VideoCaptureDescriptorPhysicalStartAddress_CH0 = pDescriptorInfo->pPhysAddr_X1_CH0;
                VideoCaptureDescriptorPhysicalStartAddress_CH1 = pDescriptorInfo->pPhysAddr_X1_CH1;
                VideoCaptureDescriptorPhysicalStartAddress_CH2 = pDescriptorInfo->pPhysAddr_X1_CH2;
                bFirstVideoCaptureBufferSubmitted = TRUE;
            }

            pXPn_CH0 = (XLLP_DMAC_DESCRIPTOR_T *)pDescriptorInfo->pVirtAddr_XPn_CH0;
            pXPn_CH1 = (XLLP_DMAC_DESCRIPTOR_T *)pDescriptorInfo->pVirtAddr_XPn_CH1;
            pXPn_CH2 = (XLLP_DMAC_DESCRIPTOR_T *)pDescriptorInfo->pVirtAddr_XPn_CH2;

            pXn_CH0 = (XLLP_DMAC_DESCRIPTOR_T *)pDescriptorInfo->pVirtAddr_Xn_CH0;
            pXn_CH1 = (XLLP_DMAC_DESCRIPTOR_T *)pDescriptorInfo->pVirtAddr_Xn_CH1;
            pXn_CH2 = (XLLP_DMAC_DESCRIPTOR_T *)pDescriptorInfo->pVirtAddr_Xn_CH2;

            pLast_Pn_CH0 = (XLLP_DMAC_DESCRIPTOR_T *)VirtAddr_Last_Pn_CH0;
            pLast_Pn_CH1 = (XLLP_DMAC_DESCRIPTOR_T *)VirtAddr_Last_Pn_CH1;
            pLast_Pn_CH2 = (XLLP_DMAC_DESCRIPTOR_T *)VirtAddr_Last_Pn_CH2;

            pLast_n_CH0 = (XLLP_DMAC_DESCRIPTOR_T *)VirtAddr_Last_n_CH0;
            pLast_n_CH1 = (XLLP_DMAC_DESCRIPTOR_T *)VirtAddr_Last_n_CH1;
            pLast_n_CH2 = (XLLP_DMAC_DESCRIPTOR_T *)VirtAddr_Last_n_CH2;

            // Chain last phantom descriptor XPn to the first phantom descriptor XP1.
            // This loops the phantom descriptor back to itself.
            pXPn_CH0->DDADR = pDescriptorInfo->pPhysAddr_XP1_CH0;
            if (isYUV_PLANAR) pXPn_CH1->DDADR = pDescriptorInfo->pPhysAddr_XP1_CH1;
            if (isYUV_PLANAR) pXPn_CH2->DDADR = pDescriptorInfo->pPhysAddr_XP1_CH2;

            // Chain the last real descriptor to the start of the phantom descriptor chain
            pXn_CH0->DDADR = pDescriptorInfo->pPhysAddr_XP1_CH0;
            if (isYUV_PLANAR) pXn_CH1->DDADR = pDescriptorInfo->pPhysAddr_XP1_CH1;
            if (isYUV_PLANAR) pXn_CH2->DDADR = pDescriptorInfo->pPhysAddr_XP1_CH2;

            // Chain the last descriptor to this new descriptor by adding this new descriptor to the end of the chain.
            // We either are already looping on the phantom descriptor by the time this frame buffer is submitted, or
            // we have yet to begin looping on the phantom descriptor.  Therefore, we must program the end of the phantom descriptor
            // to branch to the beginning of the new descriptor.  And, if we were not already looping on the phantom descriptor,
            // then we want to program the last real descriptor so that it branches to the beginning of the new descriptor.
            if (VirtAddr_Last_Pn_CH0)
            {
                pLast_Pn_CH0->DDADR = pDescriptorInfo->pPhysAddr_X1_CH0;
                if (isYUV_PLANAR) pLast_Pn_CH1->DDADR = pDescriptorInfo->pPhysAddr_X1_CH1;
                if (isYUV_PLANAR) pLast_Pn_CH2->DDADR = pDescriptorInfo->pPhysAddr_X1_CH2;

                pLast_n_CH0->DDADR = pDescriptorInfo->pPhysAddr_X1_CH0;
                if (isYUV_PLANAR) pLast_n_CH1->DDADR = pDescriptorInfo->pPhysAddr_X1_CH1;
                if (isYUV_PLANAR) pLast_n_CH2->DDADR = pDescriptorInfo->pPhysAddr_X1_CH2;
            }

            // Update pVirtAddr_Last_Pn and pVirtAddr_Last_n so they point to this new buffer
            VirtAddr_Last_Pn_CH0 = pDescriptorInfo->pVirtAddr_XPn_CH0;
            VirtAddr_Last_Pn_CH1 = pDescriptorInfo->pVirtAddr_XPn_CH1;
            VirtAddr_Last_Pn_CH2 = pDescriptorInfo->pVirtAddr_XPn_CH2;

            VirtAddr_Last_n_CH0 = pDescriptorInfo->pVirtAddr_Xn_CH0;
            VirtAddr_Last_n_CH1 = pDescriptorInfo->pVirtAddr_Xn_CH1;
            VirtAddr_Last_n_CH2 = pDescriptorInfo->pVirtAddr_Xn_CH2;

        } else if (BufferType == STILL_CAPTURE_BUFFER)
        {
          //  if (bFirstStillBufferSubmitted == FALSE)
          //  {
                StillCaptureDescriptorPhysicalStartAddress_CH0 = pDescriptorInfo->pPhysAddr_X1_CH0;
                StillCaptureDescriptorPhysicalStartAddress_CH1 = pDescriptorInfo->pPhysAddr_X1_CH1;
                StillCaptureDescriptorPhysicalStartAddress_CH2 = pDescriptorInfo->pPhysAddr_X1_CH2;
                bFirstStillBufferSubmitted = TRUE;
          //  }

                pXPn_CH0 = (XLLP_DMAC_DESCRIPTOR_T *)pDescriptorInfo->pVirtAddr_XPn_CH0;
                pXPn_CH1 = (XLLP_DMAC_DESCRIPTOR_T *)pDescriptorInfo->pVirtAddr_XPn_CH1;
                pXPn_CH2 = (XLLP_DMAC_DESCRIPTOR_T *)pDescriptorInfo->pVirtAddr_XPn_CH2;

                pXn_CH0 = (XLLP_DMAC_DESCRIPTOR_T *)pDescriptorInfo->pVirtAddr_Xn_CH0;
                pXn_CH1 = (XLLP_DMAC_DESCRIPTOR_T *)pDescriptorInfo->pVirtAddr_Xn_CH1;
                pXn_CH2 = (XLLP_DMAC_DESCRIPTOR_T *)pDescriptorInfo->pVirtAddr_Xn_CH2;

                // Chain last phantom descriptor XPn to the first phantom descriptor XP1.
                //  This loops the phantom descriptor back to itself.
                pXPn_CH0->DDADR = pDescriptorInfo->pPhysAddr_XP1_CH0;
                if (isYUV_PLANAR) pXPn_CH1->DDADR = pDescriptorInfo->pPhysAddr_XP1_CH1;
                if (isYUV_PLANAR) pXPn_CH2->DDADR = pDescriptorInfo->pPhysAddr_XP1_CH2;

                // Chain the last real descriptor to the start of the phantom descriptor chain
                pXn_CH0->DDADR = pDescriptorInfo->pPhysAddr_XP1_CH0;
                if (isYUV_PLANAR) pXn_CH1->DDADR = pDescriptorInfo->pPhysAddr_XP1_CH1;
                if (isYUV_PLANAR) pXn_CH2->DDADR = pDescriptorInfo->pPhysAddr_XP1_CH2;
        }
    } else 
    {
        if (!pBufInfo)
            NKDbgPrintfW(TEXT("Buffer is null.  Submit buffer failed.\r\n"));
        else
            NKDbgPrintfW(TEXT("Buffer list had null entry.  Submit buffer failed.  Buffer ID %d\r\n"),pBufInfo->BufferID);
    }
}

// Given a buffer allocated by the application, this function maps the buffer into discreet 4K chunks in the physical
// address space, then builds a DMA descriptor chain that performs a scatter gather DMA sequence using these 
// 4K chunks as individual target buffers.  It also builds a phantom descriptor chain which is executed after
// the primary descriptor chain completes.  The phantom descriptor chain is created so that the DMA can run
// continuously in the background, even if there is no client ready to read the data from the primary chain.  This 
// way primary buffers will remain untouched, the dma activity is undisturbed, the sensor is unaware of the flow control,
// and the primary buffers remain ready for their consumption by the application that submitted them.
// Buffers are chained together based on the order in which they are submitted by the application.  In this way,
// the application can now manage the buffer memory according to its own needs.  If an application needs to lock
// down a buffer, it simply chooses not to resubmit that buffer until it has completed using it.
int CameraPrepareBuffer(P_CAMERA_DMA_BUFFER_INFO pBufInfo, int BufferType)
{
        int limit = 0;
        int i = 0;
        int j = 0;
        int num_descriptors;
        int memory_required;
        int planes_number;

        DESCRIPTOR_INFO_T *pDescriptorInfo;
        DWORD *pPhysicalBufferAddress;
        DWORD *pPhysicalDescriptorBufferAddress;
        DWORD *VirtAddr;

        int fifo0_num_descriptors = 0;
        int fifo1_num_descriptors = 0;
        int fifo2_num_descriptors = 0;

        int fifo0_transfer_size = 0;
        int fifo1_transfer_size = 0;
        int fifo2_transfer_size = 0;
        int frame_size = (VIDEO_CAPTURE_BUFFER == BufferType ? g_HwContext.XllpCAMERA.Video_capture_width : g_HwContext.XllpCAMERA.Still_capture_width) * \
                                         (VIDEO_CAPTURE_BUFFER == BufferType ? g_HwContext.XllpCAMERA.Video_capture_height : g_HwContext.XllpCAMERA.Still_capture_height) * 2;

        unsigned int capture_onput_format;


        if (VIDEO_CAPTURE_BUFFER == BufferType)
        {
            capture_onput_format = g_HwContext.XllpCAMERA.Video_capture_output_format;
        }
        else
        {
            capture_onput_format = g_HwContext.XllpCAMERA.Still_capture_output_format;

        }

        // Determine the length of the image transfer.  This is the sum of each fifo transfer length.
        switch(capture_onput_format) 
        {
                case XLLP_CAMERA_IMAGE_FORMAT_RGB555:
                case XLLP_CAMERA_IMAGE_FORMAT_RGB565:
                        fifo0_transfer_size = frame_size;
                        planes_number = 1;
                        break;
                case XLLP_CAMERA_IMAGE_FORMAT_YCBCR422_PACKED:
                        fifo0_transfer_size = frame_size;
                        planes_number = 1;
                        break;
                case XLLP_CAMERA_IMAGE_FORMAT_YCBCR422_PLANAR:
                        fifo0_transfer_size = frame_size / 2;
                        fifo1_transfer_size = frame_size / 4;
                        fifo2_transfer_size = frame_size / 4;
                        planes_number = 3;
                        break;
                default:
                        NKDbgPrintfW(TEXT("Unsupported capture format.\r\n"));
                        return FALSE;
                        break;
        }

        // Allocate some memory to store the physical address of each descriptor used to map each 4K segment
        // We need one set of descriptors to DMA the data to a real buffer.  We need another set of descriptors
        // (phantom descriptors) to DMA the data to non-incrementing address that will be used only for flow control.
        // Therefore, twice as many descriptors are required for the whole chain, than for just the primary DMA sequence.
        // We can obtain physical addresses with 4K granularity.  And the pages are not guaranteed to be contiguous.  
        // Therefore, we'll dedicate one descriptor for every 4K of image data.

        fifo0_num_descriptors = (( fifo0_transfer_size + SINGLE_DESCRIPTOR_TRANSFER_MAX -1 )
                                 / SINGLE_DESCRIPTOR_TRANSFER_MAX) * 2;
        fifo1_num_descriptors = (( fifo1_transfer_size + SINGLE_DESCRIPTOR_TRANSFER_MAX -1 )
                                 / SINGLE_DESCRIPTOR_TRANSFER_MAX) * 2;
        fifo2_num_descriptors = (( fifo2_transfer_size + SINGLE_DESCRIPTOR_TRANSFER_MAX -1 )
                                 / SINGLE_DESCRIPTOR_TRANSFER_MAX) * 2;

        num_descriptors = fifo0_num_descriptors + fifo1_num_descriptors + fifo2_num_descriptors;
        
        memory_required = ((num_descriptors / 2) * SINGLE_DESCRIPTOR_TRANSFER_MAX);
        
        if (pBufInfo->size < memory_required)
        {
                pBufInfo->size = memory_required;
                //NKDbgPrintfW(TEXT("Memory size inadequate.  Need %d bytes\r\n"),pBufInfo->size);
                return FALSE;
        }

        if (NumBuffers >= MAX_CAMERA_DMA_BUFFERS)
        {
                NKDbgPrintfW(TEXT("Max number of buffers reached.\r\n"));
                return FALSE;
        }

        MasterBufferList[NumBuffers] = (DESCRIPTOR_INFO_T *)malloc(sizeof(DESCRIPTOR_INFO_T));
        if (MasterBufferList[NumBuffers] == NULL)
        {
                NKDbgPrintfW(TEXT("Unable to allocate memory for entry into MasterBufferList.\r\n"));
                return FALSE;
        }

        pDescriptorInfo = MasterBufferList[NumBuffers];
        memset(pDescriptorInfo, 0x0, sizeof(DESCRIPTOR_INFO_T));
        
        // Assign a unique frame ID
        pDescriptorInfo->FrameID = NumBuffers;
        pBufInfo->BufferID = NumBuffers;

        pDescriptorInfo->dwPlanesNumber = planes_number;
		VirtAddr = (DWORD *)pBufInfo->VirtAddr;



        // Set the memory to uncached, unbuffered
        VirtualSetAttributes((void *)VirtAddr, memory_required, 0, 0xC, NULL);

        pBufInfo->pY = (DWORD *)pBufInfo->VirtAddr;
        pBufInfo->pCb = (DWORD *)pBufInfo->VirtAddr + (((fifo0_num_descriptors / 2) * SINGLE_DESCRIPTOR_TRANSFER_MAX) / 4);
        pBufInfo->pCr = (DWORD *)pBufInfo->VirtAddr + (((fifo0_num_descriptors / 2) * SINGLE_DESCRIPTOR_TRANSFER_MAX) / 4)
                                                    + (((fifo1_num_descriptors / 2) * SINGLE_DESCRIPTOR_TRANSFER_MAX) / 4);

        // This list holds a pointer to each physical destination buffer
        pDescriptorInfo->pBufferPhyAddress = (DWORD *)malloc(num_descriptors * sizeof(DWORD));
        if (pDescriptorInfo->pBufferPhyAddress == NULL)
        {
                NKDbgPrintfW(TEXT("Unable to allocate memory for pDescriptorInfo->pBufferPhyAddress.\r\n"));
                return FALSE;
        }

        pDescriptorInfo->pBufferVirtAddress = VirtAddr;


        // Allocate memory for the descriptor buffer list, plus one dummy entry in order to simplify the buffer chaining logic
        // The last entry in the descriptor chain needs an entry for the DDADR register, and if we allocate an extra unused entry
        // in this list, we can assign the dummy value without any harm.
        pDescriptorInfo->pDescriptorBufferPhyAddress = (DWORD *)malloc((num_descriptors + 1) * sizeof(DWORD));
        if (!pDescriptorInfo->pDescriptorBufferPhyAddress)
        {
                NKDbgPrintfW(TEXT("Unable to allocate memory for pDescriptorInfo->pDescriptorBufferPhyAddress\r\n"));
                return FALSE;
        }


        pDescriptorInfo->pDescriptorBufferVirtAddress = (DWORD *)malloc((num_descriptors + 1) * sizeof(XLLP_DMA_DESCRIPTOR_T));
        if (!pDescriptorInfo->pDescriptorBufferVirtAddress)
        {
                NKDbgPrintfW(TEXT("Unable to allocate memory for pDescriptorInfo->pDescriptorBufferVirtAddress\r\n"));
                return FALSE;
        }

        memset(pDescriptorInfo->pDescriptorBufferVirtAddress, 0x0, (num_descriptors + 1) * sizeof(XLLP_DMA_DESCRIPTOR_T));

        // Set the descriptor memory to uncached, unbuffered so that the descriptors have valid data in them when read by the DMA engine later.
        VirtualSetAttributes((void *)pDescriptorInfo->pDescriptorBufferVirtAddress, (num_descriptors + 1) * sizeof(XLLP_DMA_DESCRIPTOR_T), 0, 0xC, NULL);

        // Create the lookup table that maps the user allocated virtual buffer address to the corresponding physical address.
        // We know that buffers are contiguous within each 4K segment.

        pPhysicalBufferAddress = pDescriptorInfo->pBufferPhyAddress;
        limit = (DWORD)VirtAddr + pBufInfo->size;
        // Now determine the physical addresses for each 4K segment and store them in the array maintained for this descriptor chain
        for (i = (DWORD)VirtAddr; i < limit; i+= SINGLE_DESCRIPTOR_TRANSFER_MAX)
        {
                getPhyAddress ((DWORD *)i, pPhysicalBufferAddress);
                //NKDbgPrintfW(TEXT("Target Buffer Virtual Address 0x%x has Physical address 0x%x\r\n"),i, *pPhysicalBufferAddress);
                pPhysicalBufferAddress++;
        }

        // Create the lookup table that maps the virtual address space that was just allocated for the descriptors to the corresponding physical address.
        pPhysicalDescriptorBufferAddress = pDescriptorInfo->pDescriptorBufferPhyAddress;        
        limit = (DWORD)(pDescriptorInfo->pDescriptorBufferVirtAddress) + ((num_descriptors + 1) * sizeof(XLLP_DMA_DESCRIPTOR_T));
        // Now determine the physical addresses for each descriptor address and store them in the array maintained for this descriptor chain
        for (i = (DWORD)(pDescriptorInfo->pDescriptorBufferVirtAddress); i < limit; i+= sizeof(XLLP_DMA_DESCRIPTOR_T))
        {
                getPhyAddress ((DWORD *)i, pPhysicalDescriptorBufferAddress);
                //NKDbgPrintfW(TEXT("Descriptor Virtual Address 0x%x has Physical address 0x%x\r\n"),i, *pPhysicalDescriptorBufferAddress);
                pPhysicalDescriptorBufferAddress++;
        }

        // Make sure we align the buffer
        if (VirtAddr_Phantom_Buffer == 0x0)
        {
                VirtAddr_Phantom_Buffer_Base_Address = (DWORD)((DWORD *)malloc(SINGLE_DESCRIPTOR_TRANSFER_MAX + 0xFFF));
                if (VirtAddr_Phantom_Buffer_Base_Address == 0x0)
                {
                        NKDbgPrintfW(TEXT("Unable to allocate memory for VirtAddr_Phantom_Buffer\r\n"));
                        return FALSE;
                }

                VirtAddr_Phantom_Buffer = VirtAddr_Phantom_Buffer_Base_Address;

                if (VirtAddr_Phantom_Buffer & 0xFFF)
                {
                        VirtAddr_Phantom_Buffer += 0x1000 - (VirtAddr_Phantom_Buffer & 0xFFF);
                }

                getPhyAddress ((DWORD *)VirtAddr_Phantom_Buffer, &((DWORD)PhysAddr_Phantom_Buffer));
                //NKDbgPrintfW(TEXT("Phantom buffer Virtual Address 0x%x has Physical address 0x%x\r\n"),VirtAddr_Phantom_Buffer, PhysAddr_Phantom_Buffer);
        }

        pDescriptorInfo->pPhysAddr_Phantom = PhysAddr_Phantom_Buffer;

        GenerateDMAChain( pDescriptorInfo, fifo0_transfer_size, fifo1_transfer_size, fifo2_transfer_size);

        NumBuffers++;

//      return (int)pDescriptorInfo->FrameID;
        return TRUE;
}

// Generate a primary descriptor chain that fills the provided buffers.
// Generate a phantom descriptor chain that executes after the primary chain.
// When buffers are chained together, each primary chain is linked to the next primary
// descriptor chain.  The last primary descriptor chain is linked to the phantom descriptor chain.
int GenerateDMAChain(
    DESCRIPTOR_INFO_T *pDescriptorInfo,
    int fifo0_transfer_size,
    int fifo1_transfer_size,
    int fifo2_transfer_size
    )
{
    XLLP_DMAC_DESCRIPTOR_T *cur_des_virtual;
    int des_transfer_size, remain_size;
    unsigned int i,j,k,l;
    unsigned int fifo0_num_descriptors = 0;
    unsigned int fifo1_num_descriptors = 0;
    unsigned int fifo2_num_descriptors = 0;

    // calculate how many descriptors are needed per frame
    fifo0_num_descriptors = (( fifo0_transfer_size + SINGLE_DESCRIPTOR_TRANSFER_MAX -1 )
                            / SINGLE_DESCRIPTOR_TRANSFER_MAX) * 2;
    fifo1_num_descriptors = (( fifo1_transfer_size + SINGLE_DESCRIPTOR_TRANSFER_MAX -1 )
                            / SINGLE_DESCRIPTOR_TRANSFER_MAX) * 2;
    fifo2_num_descriptors = (( fifo2_transfer_size + SINGLE_DESCRIPTOR_TRANSFER_MAX -1 )
                            / SINGLE_DESCRIPTOR_TRANSFER_MAX) * 2;

    cur_des_virtual = (XLLP_DMAC_DESCRIPTOR_T*)pDescriptorInfo->pDescriptorBufferVirtAddress;

    // First iteration, build the real chain.
    // Second iteration, build the phantom chain.
    k = 0;
    l = 0;
    for(i=0; i < 2; i++) {
    
        remain_size = fifo0_transfer_size;

        for(j=0; j < fifo0_num_descriptors >> 1; j++) {
            // set descriptor
            if (remain_size > SINGLE_DESCRIPTOR_TRANSFER_MAX) 
                des_transfer_size = SINGLE_DESCRIPTOR_TRANSFER_MAX;
            else
                des_transfer_size = remain_size;

            // Build the primary DMA chain
            if (i == 0)
            {
                    cur_des_virtual->DDADR = (unsigned)pDescriptorInfo->pDescriptorBufferPhyAddress[++l];
                    cur_des_virtual->DSADR = XLLP_CI_REGBASE_PHY + XLLP_CIBR0;                      // FIFO0 physical address
                    cur_des_virtual->DTADR = (unsigned)pDescriptorInfo->pBufferPhyAddress[k++];
                    cur_des_virtual->DCMD = des_transfer_size | XLLP_DMAC_DCMD_FLOW_SRC | XLLP_DMAC_DCMD_INC_TRG_ADDR | (XLLP_DMAC_BURSTSIZE_16 << 16);
        
                    if (j == 0)
                    {
                            pDescriptorInfo->pPhysAddr_X1_CH0 = (XLLP_UINT32_T)pDescriptorInfo->pDescriptorBufferPhyAddress[l-1];// Physical address of the first real descriptor
                    }
                    // Special case for the last non-phantom descriptor in the chain
                    // Enable an end of frame interrupt for this descriptor only.
                    // Save a pointer to this descriptor and reset the remaining size for the phantom descriptor chain
                    if (j == ((fifo0_num_descriptors >> 1) - 1))
                    {
                            cur_des_virtual->DCMD |= XLLP_DMAC_DCMD_END_IRQ_EN;
                            pDescriptorInfo->pVirtAddr_Xn_CH0 = (XLLP_UINT32_T)cur_des_virtual;// Virtual address of the last real descriptor
                            // Reset the DMA transfer size for the phantom descriptor
                            remain_size = fifo0_transfer_size;
                    }
            } else
            // Build the phantom DMA chain
            {
                    cur_des_virtual->DDADR = (unsigned)pDescriptorInfo->pDescriptorBufferPhyAddress[++l];
                    cur_des_virtual->DSADR = XLLP_CI_REGBASE_PHY + XLLP_CIBR0;                      // FIFO0 physical address
                    cur_des_virtual->DTADR = (unsigned)pDescriptorInfo->pPhysAddr_Phantom;
                    cur_des_virtual->DCMD = des_transfer_size | XLLP_DMAC_DCMD_FLOW_SRC | XLLP_DMAC_DCMD_INC_TRG_ADDR | (XLLP_DMAC_BURSTSIZE_16 << 16);
            
                    if (j == 0)
                    {
                            pDescriptorInfo->pPhysAddr_XP1_CH0 = (XLLP_UINT32_T)pDescriptorInfo->pDescriptorBufferPhyAddress[l-1];          // Physical address of the first phantom descriptor
                    }
                    if (j == ((fifo0_num_descriptors >> 1) - 1))
                    {
                            pDescriptorInfo->pVirtAddr_XPn_CH0 = (XLLP_UINT32_T)cur_des_virtual;    // Virtual address of the last phantom descriptor       
                    }
            }
            // advance pointers
            remain_size -= des_transfer_size;
            cur_des_virtual++;
        }
    }
    
    // generate fifo1 dma chains
    if (fifo1_transfer_size) {

        for(i=0; i < 2; i++) {
            // in each iteration, generate one dma chain for one frame
            remain_size = fifo1_transfer_size;
        
            for(j=0; j < fifo1_num_descriptors >> 1; j++) {
                // set descriptor
                if (remain_size > SINGLE_DESCRIPTOR_TRANSFER_MAX) 
                    des_transfer_size = SINGLE_DESCRIPTOR_TRANSFER_MAX;
                else
                    des_transfer_size = remain_size;

                if (i == 0)
                {
                        cur_des_virtual->DDADR = (unsigned)pDescriptorInfo->pDescriptorBufferPhyAddress[++l];
                        cur_des_virtual->DSADR = XLLP_CI_REGBASE_PHY + XLLP_CIBR1;      // FIFO1 physical address
                        cur_des_virtual->DTADR = (unsigned)pDescriptorInfo->pBufferPhyAddress[k++];
                        cur_des_virtual->DCMD = des_transfer_size | XLLP_DMAC_DCMD_FLOW_SRC | XLLP_DMAC_DCMD_INC_TRG_ADDR | (XLLP_DMAC_BURSTSIZE_8 << 16);
                
                        if (j == 0)
                                pDescriptorInfo->pPhysAddr_X1_CH1 = (XLLP_UINT32_T)pDescriptorInfo->pDescriptorBufferPhyAddress[l-1];// Physical address of the first real descriptor
                        if (j == ((fifo1_num_descriptors >> 1) - 1))
                        {
                                pDescriptorInfo->pVirtAddr_Xn_CH1 = (XLLP_UINT32_T)cur_des_virtual;// Virtual address of the last real descriptor
        
                                // Reset the DMA transfer size for the phantom descriptor
                                remain_size = fifo1_transfer_size;
                        }
                }
                else 
                {
                        cur_des_virtual->DDADR = (unsigned)pDescriptorInfo->pDescriptorBufferPhyAddress[++l];
                        cur_des_virtual->DSADR = XLLP_CI_REGBASE_PHY + XLLP_CIBR1;      // FIFO1 physical address
                        cur_des_virtual->DTADR = (unsigned)pDescriptorInfo->pPhysAddr_Phantom;
                        cur_des_virtual->DCMD = des_transfer_size | XLLP_DMAC_DCMD_FLOW_SRC | XLLP_DMAC_DCMD_INC_TRG_ADDR | (XLLP_DMAC_BURSTSIZE_8 << 16);
                        
                        if (j == 0)
                                pDescriptorInfo->pPhysAddr_XP1_CH1 = (XLLP_UINT32_T)pDescriptorInfo->pDescriptorBufferPhyAddress[l-1];          // Physical address of the first phantom descriptor
                        if (j == ((fifo1_num_descriptors >> 1) - 1))
                                pDescriptorInfo->pVirtAddr_XPn_CH1 = (XLLP_UINT32_T)cur_des_virtual;    // Virtual address of the last phantom descriptor       
                }

                // advance pointers
                remain_size -= des_transfer_size;
                cur_des_virtual++;
            }
        }
    }

    // generate fifo2 dma chains
    if (fifo2_transfer_size) {

        for(i=0; i < 2; i++) {
            // in each iteration, generate one dma chain for one frame
            remain_size = fifo2_transfer_size;
    
            for(j=0; j < fifo2_num_descriptors >> 1; j++) {
                // set descriptor
                if (remain_size > SINGLE_DESCRIPTOR_TRANSFER_MAX) 
                    des_transfer_size = SINGLE_DESCRIPTOR_TRANSFER_MAX;
                else
                    des_transfer_size = remain_size;

                if (i == 0)
                {
                        cur_des_virtual->DDADR = (unsigned)pDescriptorInfo->pDescriptorBufferPhyAddress[++l];
                        cur_des_virtual->DSADR = XLLP_CI_REGBASE_PHY + XLLP_CIBR2;      // FIFO2 physical address
                        cur_des_virtual->DTADR = (unsigned)pDescriptorInfo->pBufferPhyAddress[k++];
                        cur_des_virtual->DCMD = des_transfer_size | XLLP_DMAC_DCMD_FLOW_SRC | XLLP_DMAC_DCMD_INC_TRG_ADDR | (XLLP_DMAC_BURSTSIZE_8 << 16);
        
                        if (j == 0)
                                pDescriptorInfo->pPhysAddr_X1_CH2 = (XLLP_UINT32_T)pDescriptorInfo->pDescriptorBufferPhyAddress[l-1];// Physical address of the first real descriptor
                        if (j == ((fifo2_num_descriptors >> 1) - 1))
                        {
                                pDescriptorInfo->pVirtAddr_Xn_CH2 = (XLLP_UINT32_T)cur_des_virtual;// Virtual address of the last real descriptor
        
                                // Reset the DMA transfer size for the phantom descriptor
                                remain_size = fifo2_transfer_size;
                        }
        
                }
                else
                {
                        cur_des_virtual->DDADR = (unsigned)pDescriptorInfo->pDescriptorBufferPhyAddress[++l];
                        cur_des_virtual->DSADR = XLLP_CI_REGBASE_PHY + XLLP_CIBR2;      // FIFO2 physical address
                        cur_des_virtual->DTADR = (unsigned)pDescriptorInfo->pPhysAddr_Phantom;
                        cur_des_virtual->DCMD = des_transfer_size | XLLP_DMAC_DCMD_FLOW_SRC | XLLP_DMAC_DCMD_INC_TRG_ADDR | (XLLP_DMAC_BURSTSIZE_8 << 16);
            
                        if (j == 0)
                                pDescriptorInfo->pPhysAddr_XP1_CH2 = (XLLP_UINT32_T)pDescriptorInfo->pDescriptorBufferPhyAddress[l-1];          // Physical address of the first phantom descriptor
                        if (j == ((fifo2_num_descriptors >> 1) - 1))
                                pDescriptorInfo->pVirtAddr_XPn_CH2 = (XLLP_UINT32_T)cur_des_virtual;    // Virtual address of the last phantom descriptor       
        
                }
                // advance pointers
                remain_size -= des_transfer_size;
                cur_des_virtual++;
            }
        }
    }
    return TRUE;   
}


/***********************************************************************
 *
 * Private functions
 *
 ***********************************************************************/
void PrvStartDMATransfer(P_HW_CONTEXT_T pHwContext, unsigned block_id)
{
    XLLP_DMAC_DESCRIPTOR_T *des_virtual, *des_physical;


    des_virtual = (P_XLLP_DMAC_DESCRIPTOR_T)NULL;
    des_physical = (P_XLLP_DMAC_DESCRIPTOR_T)pHwContext->XllpDMA.fifo0_descriptors_physical;
    XllpDmacCfgChannelDescTransfer( //des_virtual,
                               des_physical,
                               pHwContext->XllpDMA.dma_channels[0],        
                               CAMERA_DMA_DEVICE[0],                    
                               XLLP_DMAC_ALIGNMENT_OFF              
                              );

    if ( pHwContext->XllpDMA.fifo1_transfer_size ) {
                des_virtual = (P_XLLP_DMAC_DESCRIPTOR_T)NULL;
                des_physical = (P_XLLP_DMAC_DESCRIPTOR_T)pHwContext->XllpDMA.fifo1_descriptors_physical;
                XllpDmacCfgChannelDescTransfer( //des_virtual,
                                   des_physical,
                                   pHwContext->XllpDMA.dma_channels[1],        
                                   CAMERA_DMA_DEVICE[1],                    
                                   XLLP_DMAC_ALIGNMENT_OFF              
                                  );
        }

    if ( pHwContext->XllpDMA.fifo2_transfer_size ) {
        des_virtual = (P_XLLP_DMAC_DESCRIPTOR_T)NULL;
        des_physical = (P_XLLP_DMAC_DESCRIPTOR_T)pHwContext->XllpDMA.fifo2_descriptors_physical;
        XllpDmacCfgChannelDescTransfer( //des_virtual,
                                   des_physical,
                                   pHwContext->XllpDMA.dma_channels[2],        
                                   CAMERA_DMA_DEVICE[2],                    
                                   XLLP_DMAC_ALIGNMENT_OFF              
                                  );
    }

    if ( pHwContext->XllpDMA.fifo0_transfer_size &&
         pHwContext->XllpDMA.fifo1_transfer_size &&
         pHwContext->XllpDMA.fifo2_transfer_size ) 
    {
        XllpDmacStartTransfer( pHwContext->XllpDMA.dma_channels[0] );
        XllpDmacStartTransfer( pHwContext->XllpDMA.dma_channels[1] );
        XllpDmacStartTransfer( pHwContext->XllpDMA.dma_channels[2] );
    } else 
    {
        XllpDmacStartTransfer( pHwContext->XllpDMA.dma_channels[0] );
    }

}

void PrvStopDMATransfer(P_HW_CONTEXT_T pHwContext)
{
    unsigned int capture_output_format = pHwContext->XllpCAMERA.Video_capture_output_format;
    if (pHwContext->XllpCAMERA.capture_mode == XLLP_CAMERA_MODE_STILL)
    {
       capture_output_format = pHwContext->XllpCAMERA.Still_capture_output_format;
    }

    // stop channel 0
    XllpDmacStopTransfer( pHwContext->XllpDMA.dma_channels[0] );

    // stop channel 1 and 2  
    if ( capture_output_format == XLLP_CAMERA_IMAGE_FORMAT_YCBCR422_PLANAR ) 
    {
        XllpDmacStopTransfer( pHwContext->XllpDMA.dma_channels[1] );
        XllpDmacStopTransfer( pHwContext->XllpDMA.dma_channels[2] );
    }
}

XLLP_STATUS_T PrvStartCapture(P_HW_CONTEXT_T pHwContext, unsigned int block_id, unsigned int frames)
{
    XLLP_STATUS_T   status;
    
    // clear ci fifo
    XllpCIResetFIFO(pHwContext->XllpCAMERA.ci_reg_base);
    XllpCIClearInterruptStatus(pHwContext->XllpCAMERA.ci_reg_base, 0xFFFFFFFF);

    // start dma
    PrvStartDMATransfer(pHwContext, block_id);

    // start capture
    status = pHwContext->XllpCAMERA.camera_functions->start_capture(&pHwContext->XllpCAMERA, frames);    
    return status;
}


// Set the image format
XLLP_STATUS_T PrvSetDMARingBuffer(P_HW_CONTEXT_T pHwContext)
{
    unsigned int frame_size;
    unsigned int capture_output_format = pHwContext->XllpCAMERA.Video_capture_output_format;
    if (pHwContext->XllpCAMERA.capture_mode == XLLP_CAMERA_MODE_STILL)
    {
       capture_output_format = pHwContext->XllpCAMERA.Still_capture_output_format;
    }

    // ring buffer init
    switch(capture_output_format) {
        case XLLP_CAMERA_IMAGE_FORMAT_RGB555:
        case XLLP_CAMERA_IMAGE_FORMAT_RGB565:
            frame_size = pHwContext->XllpCAMERA.capture_width * pHwContext->XllpCAMERA.capture_height * 2;
            pHwContext->XllpDMA.fifo0_transfer_size = frame_size;
            pHwContext->XllpDMA.fifo1_transfer_size = 0;
            pHwContext->XllpDMA.fifo2_transfer_size = 0;
            break;
        case XLLP_CAMERA_IMAGE_FORMAT_YCBCR422_PACKED:
            frame_size = pHwContext->XllpCAMERA.capture_width * pHwContext->XllpCAMERA.capture_height * 2;
            pHwContext->XllpDMA.fifo0_transfer_size = frame_size;
            pHwContext->XllpDMA.fifo1_transfer_size = 0;
            pHwContext->XllpDMA.fifo2_transfer_size = 0;
            break;
        case XLLP_CAMERA_IMAGE_FORMAT_YCBCR422_PLANAR:
            frame_size = pHwContext->XllpCAMERA.capture_width * pHwContext->XllpCAMERA.capture_height * 2;
            pHwContext->XllpDMA.fifo0_transfer_size = frame_size / 2;
            pHwContext->XllpDMA.fifo1_transfer_size = frame_size / 4;
            pHwContext->XllpDMA.fifo2_transfer_size = frame_size / 4;
            break;
        default:
            return XLLP_STATUS_WRONG_PARAMETER;
            break;
    }

    pHwContext->XllpDMA.block_size = frame_size;
    pHwContext->XllpDMA.block_number = pHwContext->XllpDMA.buf_size / frame_size;
    pHwContext->XllpDMA.block_header = pHwContext->XllpDMA.block_tail = 0;
    
    return XLLP_STATUS_SUCCESS;
}

// Set the image format
XLLP_STATUS_T PrvCameraSetCaptureFormat(P_HW_CONTEXT_T pHwContext)
{
    XLLP_STATUS_T status;

    status = XllpCameraSetCaptureFormat(&pHwContext->XllpCAMERA);

    if(status == XLLP_STATUS_SUCCESS)
    {
        status = PrvSetDMARingBuffer(pHwContext);
    }

    return status;
}


int  CameraSleep(void)
{
    RETAILMSG(1, (TEXT("CameraSleep\r\n")));

    return 0;
}

int  CameraResume(void)
{
    RETAILMSG(1, (TEXT("CameraResume\r\n")));

    XllpI2cInit((P_XLLP_I2C_T)(v_pI2C), (P_XLLP_GPIO_T) v_pGPIORegs, (P_XLLP_CLKMGR_T) v_pClkRegs, (XLLP_UINT32_T) 0);

    CameraInitSensor();

    return 0;
}
